from os.path import exists


def next_path(path_pattern):
    b = get_filename_num(path_pattern)
    return path_pattern % (b + 1)


def last_path(path_pattern):
    b = get_filename_num(path_pattern)
    return path_pattern % b


def get_filename_num(path_pattern):
    """
    Finds the next free path in an sequentially named list of files

    e.g. path_pattern = 'file-%s.txt':

    file-1.txt
    file-2.txt
    file-3.txt

    Runs in log(n) time where n is the number of existing files in sequence
    """
    i = 1

    # First do an exponential search
    while exists(path_pattern % i):
        i = i * 2

    # Result lies somewhere in the interval (i/2..i]
    # We call this interval (a..b] and narrow it down until a + 1 = b
    a, b = (i // 2, i)
    while a + 1 < b:
        c = (a + b) // 2  # interval midpoint
        a, b = (c, b) if exists(path_pattern % c) else (a, c)

    return b - 1



from pprint import pprint as pp

from polly.sources.mongo.utils.traverser import DocumentTraverser
from polly.utils import Tokenizer


class MongoIter:
    def __init__(
        self,
        database_table_columns,
        database_name,
        database_client,
        database_cursor_batch_size=10000,
        **kwargs,
    ):
        self.database_name = database_name
        self.database_client = database_client
        self.database_table_columns = database_table_columns
        self.table_hash = self._get_indexable_schema_elements()
        self.limit_per_collection = kwargs.get("limit_per_table", 0)
        self.tokenizer = kwargs.get("tokenizer", Tokenizer())
        self.database_cursor_batch_size = database_cursor_batch_size

    def _schema_element_validator(self, table, column):
        return True

    def _get_indexable_schema_elements(self):

        table_hash = {}
        for table, column in self.database_table_columns:
            table_hash.setdefault(table, []).append(column)

        print("Indexable attributes")
        pp(table_hash)

        return table_hash

    def __iter__(self):
        collection_count = 0
        for collection, attributes in self.table_hash.items():
            collection_count += 1

            indexable_attributes = [
                attr
                for attr in attributes
                if self._schema_element_validator(collection, attr)
            ]

            if len(indexable_attributes) == 0:
                continue

            print(
                f"{collection_count} of {len(self.table_hash.keys())} collections indexed"
            )

            database = self.database_client[self.database_name]
            collection_document_count = database[collection].count_documents({})
            document_count = 0

            traverser = DocumentTraverser()

            for document in database[collection].find(
                projection=attributes,
                batch_size=self.database_cursor_batch_size,
                show_record_id=True,
            ):

                document_record_id = document["$recordId"]
                del document["$recordId"]

                document_count += 1

                if (document_count % self.database_cursor_batch_size) == 0:
                    print(
                        f"Indexed {document_count} of {collection_document_count} for collection {collection}"
                    )

                traverser.traverse(document)
                indexable_content = traverser.get_indexable_content()

                for column, content in indexable_content:
                    content = str(content)
                    # print('Content for tokenization (Column {}): {}'.format(column, content))

                    tokens = self.tokenizer.tokenize(content)

                    for token in tokens:
                        yield collection, document_record_id, column, token

                traverser.cleanup()



class Graph:
    __slots__ = ["_graph_dict", "_edges_info"]

    def __init__(self, graph_dict=None, edges_info=None):
        if graph_dict is None:
            graph_dict = {}
        self._graph_dict = graph_dict

        if edges_info is None:
            edges_info = {}
        self._edges_info = edges_info

    def add_vertex(self, vertex):
        self._graph_dict.setdefault(vertex, (set(), set()))
        return vertex

    def add_outgoing_neighbor(self, vertex, neighbor):
        self._graph_dict[vertex][0].add(neighbor)

    def add_incoming_neighbor(self, vertex, neighbor):
        self._graph_dict[vertex][1].add(neighbor)

    def outgoing_neighbors(self, vertex):
        yield from self._graph_dict[vertex][0]

    def incoming_neighbors(self, vertex):
        yield from self._graph_dict[vertex][1]

    def neighbors(self, vertex):
        # This method does not use directed_neighbors to avoid generating tuples with directions
        yield from self.outgoing_neighbors(vertex)
        yield from self.incoming_neighbors(vertex)

    def directed_neighbors(self, vertex):
        # This method does not use directed_neighbors to avoid generating tuples with directions
        for outgoing_neighbor in self.outgoing_neighbors(vertex):
            yield (">", outgoing_neighbor)
        for incoming_neighbor in self.incoming_neighbors(vertex):
            yield ("<", incoming_neighbor)

    def add_edge(self, vertex1, vertex2, edge_info=None, edge_direction=">"):
        if edge_direction == ">":
            self.add_outgoing_neighbor(vertex1, vertex2)
            self.add_incoming_neighbor(vertex2, vertex1)
            if edge_info is not None:
                self._edges_info[(vertex1, vertex2)] = edge_info

        elif edge_direction == "<":
            self.add_edge(vertex2, vertex1, edge_info=edge_info)
        else:
            raise SyntaxError("edge_direction must be > or <")

    def get_edge_info(self, vertex1, vertex2):
        return self._edges_info.get((vertex1, vertex2), None)

    def vertices(self):
        return self._graph_dict.keys()

    def edges(self):
        for vertex in self.vertices():
            for neighbor in self.outgoing_neighbors(vertex):
                yield (vertex,) + (neighbor,)

    def dfs_pair_iter(
        self,
        source_iter=None,
        root_predecessor=False,
        rev=False,
        **kwargs,
    ):
        last_vertex_by_level = []

        if source_iter is None:
            source_iter = self.leveled_dfs_iter(rev=rev, **kwargs)

        for direction, level, vertex in source_iter:
            if level < len(last_vertex_by_level):
                last_vertex_by_level[level] = vertex
            else:
                last_vertex_by_level.append(vertex)

            if level > 0:
                prev_vertex = last_vertex_by_level[level - 1]
                yield (prev_vertex, direction, vertex)
            elif root_predecessor:
                yield (None, "", vertex)

    def leveled_dfs_iter(
        self,
        start_vertex=None,
        visited=None,
        level=0,
        direction="",
        two_way_transversal=True,
        rev=False,
    ):
        if len(self) > 0:
            if start_vertex is None:
                start_vertex = self.get_starting_vertex(rev)
            if visited is None:
                visited = set()
            visited.add(start_vertex)

            yield ((direction, level, start_vertex))

            for neighbor in self.outgoing_neighbors(start_vertex):
                if neighbor not in visited:
                    yield from self.leveled_dfs_iter(
                        neighbor,
                        visited,
                        level=level + 1,
                        direction=">",
                        two_way_transversal=two_way_transversal,
                        rev=rev,
                    )

            # two_way_transversal indicates whether the DFS will expand through incoming neighbors
            if two_way_transversal:
                for neighbor in self.incoming_neighbors(start_vertex):
                    if neighbor not in visited:
                        yield from self.leveled_dfs_iter(
                            neighbor,
                            visited,
                            level=level + 1,
                            direction="<",
                            two_way_transversal=two_way_transversal,
                            rev=rev,
                        )

    def leaves(self):
        for vertice in self.vertices():
            if sum(1 for neighbor in self.neighbors(vertice)) == 1:
                yield vertice

    def num_leaves(self):
        return len(list(self.leaves()))

    def get_starting_vertex(self, rev):
        vertices = iter(self.vertices())

        if rev:
            return vertices[-1]

        return next(vertices)

    def __repr__(self):
        return repr(self._graph_dict)

    def __len__(self):
        return len(self._graph_dict)

    def str_graph_dict(self):
        return str(self._graph_dict)

    def str_edges_info(self):
        return str(self._edges_info)



import pickle
from collections import namedtuple
from json import load as json_load
from pathlib import Path


class DictToSTupleFormatter(object):

    _NAME = "STuple"

    def format(self, dictionary):
        sorted_keys = sorted(dictionary.keys())
        sorted_values = []

        for key in sorted_keys:
            sorted_values.append(dictionary[key])

        formatted_keys = [
            k.split(".")[1] if "." in k else k for k in sorted_keys
        ]

        try:
            stuple = namedtuple(
                self._NAME,
                formatted_keys,
            )
        except ValueError as e:
            keys_set = set([])
            for key in formatted_keys:
                keys_set.add(key)
            stuple = namedtuple(
                self._NAME,
                keys_set,
            )
            sorted_values = []
            for key in keys_set:
                sorted_values.append(dictionary[key])

        generated_stuple = stuple(
            *sorted_values,
        )

        return generated_stuple



from os import walk
from pprint import pprint as pp

import polars as plrs
from pylathedb.handlers.base import Context, Handler
from pylathedb.index.schema_index import SchemaIndex
from pylathedb.index.type_index import TypeIndex

from ...utils.tokenizer import Tokenizer
from .csv_iter import CSVIter


class CSVHandler(Handler):
    def __init__(self, config):
        self.config = config

    def get_data_paths_and_attributes(self, file_path: str):
        """
        We must receive the path that contains the files we need to retrieve and
        index. Then, we walk recursively inside the directory reading all the metadata
        (file name, column names from csv) and listing for later, when we will
        index the file contents.
        """

        data_paths_attributes = {}

        for dirpath, _, filenames in walk(file_path):
            # print(f'Current path: {dirpath}')

            if len(filenames):
                for filename in filenames:
                    file = dirpath + "/" + filename
                    # print(f'Indexing file: {file}')

                    # Read header
                    header = plrs.scan_csv(file).columns

                    data_paths_attributes[file] = header

        return data_paths_attributes

    def iterate_over_keywords(self, schema_index: SchemaIndex, **kwargs):
        print("Iterating over keywords")
        data_paths_attributes = schema_index.tables_attributes()
        # print(f'Data Paths Attributes: {data_paths_attributes}')

        return CSVIter(
            self.config, data_paths_attributes, Tokenizer(), **kwargs
        ).iterate()

    def get_fk_constraints(self):
        return {}

    def get_column_types(self, file_path: str, type_index: TypeIndex):
        for dirpath, _, filenames in walk(file_path):
            if len(filenames):
                for filename in filenames:
                    print(f"Generating type index for: {filename}")
                    file = dirpath + "/" + filename
                    table_name = file.split("/")[-1].replace(".csv", "")
                    type_index[table_name] = {}

                    # Read header
                    header = plrs.scan_csv(file).columns
                    schema_types = {}
                    # for column in header:
                    #     schema_types[column] = plrs.String

                    df = plrs.read_csv(file)  # , schema=schema_types)
                    for column in header:
                        print(
                            f"Indexing table {file} with column {column} and type: {df[column].dtype}"
                        )
                        type_index[table_name][column] = df[column].dtype


if __name__ == "__main__":
    c = Context(CSVHandler())
    result = c.get_data_paths_and_attributes("./files")
    pp(result)



import json

from polly.sources.enum import DataSourceType
from polly.sources.mongo.handler import MongoHandler
from polly.sources.postgres.handler import PostgresHandler
from polly.sources.utils.interfaces import ConnectionConfig


class SourceHandler:
    __slots__ = [
        "name",
        "type",
        "connection_config",
        "connection",
        "database",
        "_handler",
    ]

    def __init__(
        self,
        name: str,
        type: DataSourceType,
        connection_config: ConnectionConfig = None,
    ):
        self.name: str = name
        self.type: DataSourceType = type
        self.connection_config: ConnectionConfig = connection_config
        self.database: str = connection_config.database
        self._handler = None

        self.init_connection()

    def init_connection(self):
        if self.type == DataSourceType.RELATIONAL:
            self.connection = self._init_relational_connection()
        elif self.type == DataSourceType.DOCUMENT:
            self.connection = self._init_mongo_connection()
        elif self.type == DataSourceType.CSV:
            pass

    def _init_relational_connection(self):
        self._handler = PostgresHandler(
            host=self.connection_config.host,
            port=self.connection_config.port,
            user=self.connection_config.user,
            password=self.connection_config.password,
            database=self.connection_config.database,
        )

    def _init_mongo_connection(self):
        self._handler = MongoHandler(
            host=self.connection_config.host,
            port=self.connection_config.port,
            user=self.connection_config.user,
            password=self.connection_config.password,
            database=self.connection_config.database,
        )

    def _init_csv_connection(self):
        pass

    def get_fk_constraints(self):
        return self._handler.get_fk_constraints()

    def exist_results(self, statement: str):
        return self._handler.exist_results(statement)

    def get_collections_names(self):
        return self._handler.get_collections_names(self.database)

    def get_collections_and_attributes(self):
        indexable_attributes = {}
        attributes_filepath = (
            "./indexable_dataset_attributes/{}_attributes.json".format(
                self.database,
            )
        )

        with open(
            attributes_filepath,
            "r",
        ) as indexable_attributes_file:
            content = json.load(indexable_attributes_file)

            for item in content:
                indexable_attributes[item["table"]] = item["attributes"]

        return self._handler.get_collections_and_attributes(
            self.database, indexable_attributes=indexable_attributes
        )

    def execute_query(self, query):
        result = self._handler.execute(query)

        return result

    def get_description_example_values(self):
        return self._handler.get_description_example_values()



import itertools

from nltk.corpus import wordnet as wn
from nltk.stem import WordNetLemmatizer


class Similarity:
    def __init__(self, schema_index, **kwargs):
        self.schema_index = schema_index

        self.use_path_sim = kwargs.get("use_path_sim", True)
        self.use_wup_sim = kwargs.get("use_wup_sim", True)
        self.use_jaccard_sim = kwargs.get("use_jaccard_sim", True)
        self.lemmatizer = WordNetLemmatizer()

        # aggregation_method can be 'max' or 'avg'
        self.aggregation_method = kwargs.get("aggregation_method", "max")

    def path_similarity(self, word_a, word_b):
        A = set(wn.synsets(word_a))
        B = set(wn.synsets(word_b))

        path_similarities = [0]

        for sense1, sense2 in itertools.product(A, B):
            path_similarities.append(wn.path_similarity(sense1, sense2) or 0)

        return max(path_similarities)

    def wup_similarity(self, word_a, word_b):
        A = set(wn.synsets(word_a))
        B = set(wn.synsets(word_b))

        wup_similarities = [0]

        for sense1, sense2 in itertools.product(A, B):
            wup_similarities.append(wn.wup_similarity(sense1, sense2) or 0)

        return max(wup_similarities)

    def jaccard_similarity(self, word_a, word_b):

        A = set(word_a)
        B = set(word_b)

        return (len(A & B) * 1.0) / (len(A | B) * 1.0)

    def word_similarity(self, word, table, column="*"):
        sim_list = []

        lemmatized_word = self.lemmatizer.lemmatize(word)

        if column == "*":
            schema_term = table
        else:
            schema_term = column

        if self.use_path_sim:
            sim_list.append(self.path_similarity(schema_term, lemmatized_word))

        if self.use_wup_sim:
            sim_list.append(self.wup_similarity(schema_term, lemmatized_word))

        if self.use_jaccard_sim:
            sim_list.append(
                self.jaccard_similarity(schema_term, lemmatized_word)
            )

        if len(sim_list) == 0:
            return 0

        if self.aggregation_method == "max":
            return max(sim_list)
        elif self.aggregation_method == "avg":
            return (sum(sim_list) * 1.0) / (len(sim_list) * 1.0)



class KeywordQuery(object):

    __slots__ = ["keyword_query", "parsed_keyword_query", "keywords", "length"]

    def __init__(self, parsed_keyword_query, keyword_query):
        self.parsed_keyword_query = parsed_keyword_query
        self.keyword_query = keyword_query

    def get_value(self):
        return self.keyword_query

    def get_parsed_value(self):
        return self.parsed_keyword_query

    def get_keywords(self):

        if hasattr(self, "keywords"):
            return self.keywords

        keywords = []

        for element in self.parsed_keyword_query:
            if isinstance(element, list):
                for item in element:
                    keywords.append(item.replace('"', ""))
                continue

            keywords.append(element.replace('"', ""))

        self.keywords = keywords
        self.length = len(keywords)

        return keywords



import keyword
from pathlib import Path
from typing import Dict, List

from polly.core.classes.keyword_match import KeywordMatch
from polly.llm.chat import LLMChat


class LLMKMatcherResult:
    keyword_match_json_data: List[Dict[str, str]]
    keyword_matches: List[KeywordMatch]

    def __init__(
        self, data: List[Dict[str, str]], keyword_matches: List[KeywordMatch]
    ):
        self.keyword_match_json_data = data
        self.keyword_matches = keyword_matches


class LLMKMatcher:
    __slots__ = [
        "keyword_query",
        "database",
        "llm_chat",
        "model",
        "temperature",
        "top_p",
        "prompt_template",
    ]

    def __init__(
        self,
        keyword_query: str,
        database: str,
        llm_chat: LLMChat,
    ):
        self.keyword_query = keyword_query
        self.database = database
        self.llm_chat = llm_chat

    def generate_keyword_matches(self) -> List[KeywordMatch]:
        system_prompt = self._get_system_prompt()
        system_message = self.llm_chat.build_system_message(system_prompt)

        user_prompt = self._generate_user_prompt()
        user_message = self.llm_chat.build_user_message(user_prompt)

        messages = [system_message, user_message]

        keyword_match_json_data = self.llm_chat.chat(
            messages,
            cast_response_to_json=True,
        )

        keyword_matches = self._parse_keyword_matches(keyword_match_json_data)

        result = LLMKMatcherResult(keyword_match_json_data, keyword_matches)

        return result

    def _parse_keyword_matches(
        self, keyword_match_json_data: List[Dict[str, str]]
    ) -> List[KeywordMatch]:
        kms = []

        for keyword_match_json in keyword_match_json_data:
            km = KeywordMatch(
                table=keyword_match_json["table"],
                database=self.database,
            )

            if keyword_match_json["type"] == "schema":
                km.add_schema_filter(
                    attribute=keyword_match_json["column"],
                    keywords=[keyword_match_json["keyword"]],
                )
            elif keyword_match_json["type"] == "value":
                km.add_value_filter(
                    attribute=keyword_match_json["column"],
                    keywords=[keyword_match_json["keyword"]],
                )

            kms.append(km)

        return kms

    def _get_system_prompt(self):
        system_prompt = f"Below are the tables from the {self.database} database with their column names and descriptions in JSON format: \n"

        schema_description_filepath = f"tmp/{self.database}.json"

        # print(f"Opening {schema_description_filepath}")

        if not Path(schema_description_filepath).exists():
            raise Exception(
                f"Schema description file not found: {schema_description_filepath}"
            )

        with open(schema_description_filepath, "r") as file:
            system_prompt += file.read()

        return system_prompt

    def _generate_user_prompt(self):
        user_prompt = f"""
            Task: Map each keyword from the query to the most relevant table(s) and column(s) in the {self.database} database, based on the description of the table(s) and column(s).

            Keywords query: {self.keyword_query}

            Instructions:
            1. For each keyword in the query, analyze its meaning and determine the most relevant table(s) and column(s) from the {self.database} database based on the table and column descriptions.
            2. A keyword may relate to multiple tables and columns, so ensure to map it to all relevant ones.
            3. If a keyword cannot be mapped to any table or column, assign "None" to both the table and column.
            4. A keyword can be mapped to a table's name, column's name or column's value. In the case where a keyword is mapped to a table's name, indicate it using an attribute called "self", as shown below.
            5. Return the mapping in plain text using the JSON schema below. Do not output markdown. 

            Final output only according to the example json schema below in plain text:
            [
                {{"keyword": "population", "table": "country", "column": "population", "type": "schema"}},
                {{"keyword": "country", "table": "country", "column": "self", "type": "schema"}},
                {{"keyword": "texas", "table": "city", "column": "name", "type": "value"}},
                {{"keyword": "usa", "table": "country", "column": "name", "type": "value"}},
            ]
        """

        return user_prompt



class AttributeTreeGenerator(dict):
    __slots__ = ["_dict", "_single_length_attribute_tree"]

    def __init__(self):
        self._dict = {}
        self._single_length_attribute_tree = []

    def add_attribute(self, attribute, value):
        attribute_path = attribute.split(".")
        if len(attribute_path) == 1:
            self._single_length_attribute_tree.append(
                (
                    "",
                    [
                        (
                            attribute,
                            value,
                        )
                    ],
                )
            )
        else:
            leaf_attribute = attribute_path[-1]
            path = ".".join(
                attribute_path[:-1],
            )

            if path not in self._dict:
                self._dict[path] = [
                    (leaf_attribute, value),
                ]
            else:
                self._dict[path].append(
                    (
                        leaf_attribute,
                        value,
                    )
                )

    def generate(self):
        attribute_trees = []
        for root in self._dict:
            tree = (
                root,
                self._dict[root],
            )
            attribute_trees.append(tree)

        for tree in self._single_length_attribute_tree:
            attribute_trees.append(tree)

        return attribute_trees

    def __repr__(self):
        return self._dict



import json
from random import sample
from typing import List

from polly.core.classes.keyword_match import KeywordMatch

"""
Query match is a set of tuple-sets that, if properly joined,
can produce networks of tuples that fulfill the query. They
can be thought as the leaves of a Candidate Network.
"""


class QueryMatch:

    def __init__(self, keyword_matches: List[KeywordMatch]):
        self.keyword_matches = keyword_matches
        # self.value_score = 1.0
        # self.schema_score = 1.0
        # self.total_score = 1.0
        # self.tables_on_match = set()

    def get_km_from_keyword(self, keyword):
        # This is still somewhat random
        for keyword_match in self.keyword_matches:
            if keyword in keyword_match.keywords():
                return keyword_match
        return None

    def get_random_keyword_match(self):
        return sample(self.keyword_matches, k=1)[0]

    def calculate_total_score(
        self, value_index, schema_index, similarity, weight_scheme
    ):
        self.total_score = 1

        has_value_terms = self.calculate_value_score(
            value_index, schema_index, weight_scheme
        )
        has_schema_terms = self.calculate_schema_score(similarity)

        if has_value_terms:
            self.total_score *= self.value_score

        if has_schema_terms:
            self.total_score *= self.schema_score

        self.total_score /= len(self)

        # logger.info("scores for : {} value_score: {} schema_score: {} tables on match: {} total: {}".format(self.keyword_matches,
        # self.value_score,
        # self.schema_score,
        # len(self.tables_on_match),
        # self.total_score))

    def calculate_schema_score(self, similarity):
        self.schema_score = 1
        has_schema_terms = False
        for keyword_match in self.keyword_matches:
            for (
                table,
                attribute,
                schema_words,
            ) in keyword_match.schema_mappings():
                has_schema_terms = True
                sim_values = [
                    similarity.word_similarity(term, table, attribute)
                    for term in schema_words
                ]
                self.schema_score *= sum(sim_values) / len(sim_values)
        return has_schema_terms

    def calculate_value_score(self, value_index, schema_index, weight_scheme):
        has_value_terms = False
        self.value_score = 1

        for keyword_match in self.keyword_matches:
            for table, attribute, keywords in keyword_match.value_mappings():
                has_value_terms = True
                metrics = schema_index["./exported/mondial/" + table + ".csv"][
                    attribute
                ]
                tf_iaf_sum = 0
                # self.tables_on_match.add(table)
                for term in keywords:
                    tf = value_index.get_tf(
                        weight_scheme,
                        term,
                        "./exported/mondial/" + table + ".csv",
                        attribute,
                        metrics["max_frequency"],
                    )
                    iaf = value_index.get_iaf(weight_scheme, term)
                    tf_iaf_sum += tf * iaf

                cos = tf_iaf_sum / metrics["norm"][weight_scheme]
                self.value_score *= cos
        return has_value_terms

    def __eq__(self, other):
        return (
            isinstance(other, QueryMatch)
            and self.keyword_matches == other.keyword_matches
        )

    def __iter__(self):
        return iter(self.keyword_matches)

    def __len__(self):
        return len(self.keyword_matches)

    def __repr__(self):
        return repr(self.keyword_matches)

    def __str__(self):
        return repr(self.keyword_matches)

    def to_json_serializable(self):
        return [
            keyword_match.to_json_serializable()
            for keyword_match in self.keyword_matches
        ]

    def to_json(self):
        return json.dumps(self.to_json_serializable())

    @staticmethod
    def from_json_serializable(json_serializable_qm, database: str):
        return QueryMatch(
            {
                KeywordMatch.from_json_serializable(
                    json_serializable_km, database
                )
                for json_serializable_km in json_serializable_qm
            }
        )

    @staticmethod
    def from_json(str_json):
        return QueryMatch.from_json_serializable(json.loads(str_json))



import pickle


class MongoQueryBuilder:

    PROJECT_STEP_FIELD = "$project"
    MATCH_STEP_FIELD = "$match"
    LOOKUP_STEP_FIELD = "$lookup"
    ADD_FIELDS_STEP_FIELD = "$addFields"
    FILTER_STEP_FIELD = "$filter"
    COUNT_STEP_FIELD = "$count"
    UNWIND_STEP_FIELD = "$unwind"
    SET_STEP_FIELD = "$set"
    EXPRESSION_FIELD = "$expr"
    GROUP_STEP_FIELD = "$group"
    FIRST_FIELD = "$first"

    def __init__(self, database_name):
        self.query_steps = []
        self._last_added_step = None
        self._starting_collection = None
        self._last_collection_lookup = None
        self.database_name = database_name
        self.collection_structure = self._load_database_structure(
            self.database_name,
        )
        self._top_level_attributes = self._extract_top_level_attributes()

    def set_starting_collection(self, collection):
        self._starting_collection = collection

    def get_starting_collection(self):
        return self._starting_collection

    def _load_database_structure(self, collection):
        collection_structure_path = "".join(
            [
                "tmp/",
                collection + ".pickle",
            ]
        )

        with open(collection_structure_path, "rb") as f:
            collection_structure = pickle.load(
                f,
            )

        return collection_structure

    def _extract_top_level_attributes(self):
        top_level_attributes = []
        for attribute in self.collection_structure:
            if "." not in attribute:
                top_level_attributes.append(
                    attribute,
                )

    def project(self, fields):
        project_fields = {}
        if type(fields) == list:
            for field in fields:
                project_fields[field] = 1

        project_clause = {
            self.PROJECT_STEP_FIELD: project_fields,
        }

        self.query_steps.append(project_clause)

    def _squash_project(self, fields):
        for field in fields:
            self.query_steps[-1][field] = 1

    def _regex_match(self, field_name, field_value, from_filter=None):
        if from_filter:
            input_name = field_name
        else:
            input_name = "".join(["$", field_name])
        return {
            "$regexMatch": {
                "input": input_name,
                "regex": field_value,
                "options": "i",
            }
        }

    def _group(self, collection):

        group_step = {
            self.GROUP_STEP_FIELD: {
                "_id": self._to_mongo_attribute_name("_id"),
            },
        }

        for path in self.collection_structure[collection]:
            if "." not in path:
                group_step[self.GROUP_STEP_FIELD][path] = {
                    self.FIRST_FIELD: self._to_mongo_attribute_name(path)
                }

        self.query_steps.append(group_step)

    def _get_aliased_path(self, field_name, aliased_dict):
        aliased_path = field_name
        for aliased_field, aliased_field_name in aliased_dict.items():
            aliased_path = aliased_path.replace(
                aliased_field,
                aliased_field_name,
            )

        return aliased_path

    def _append_attribute_to_path(self, path, attribute):
        if attribute:
            return ".".join([path, attribute])
        return attribute

    def _get_attribute_type(self, collection, attribute):
        return self.collection_structure[collection][attribute]

    def new_match_v3(self, collection, attribute_matches):
        alias_dict = {}

        for attribute_path, matches in attribute_matches:
            # print('Attribute path: {} Matches: {}'.format(attribute_path, matches))
            path = attribute_path.split(".")
            last_attribute = None
            unwind_stage = []

            for attribute in path:

                if not attribute:
                    continue

                if last_attribute:
                    attribute = self._append_attribute_to_path(
                        last_attribute,
                        attribute,
                    )

                attribute_type = self._get_attribute_type(
                    collection,
                    attribute,
                )

                if attribute_type == list:
                    if (
                        not unwind_stage
                        and not self._check_for_list_in_parent_attributes(
                            collection, attribute, path, alias_dict
                        )
                    ):
                        # print('Copying field: {}'.format(attribute))
                        field_name, set_stage = self._copy_field(
                            attribute,
                        )

                        unwind_stage.append(
                            set_stage,
                        )

                        alias_dict[attribute] = field_name

                    aliased_attribute = self._get_aliased_path(
                        attribute,
                        alias_dict,
                    )

                    unwind_stage.append(
                        self.unwind(
                            aliased_attribute,
                        )
                    )

                last_attribute = attribute

            for attribute, value in matches:
                if last_attribute:
                    attribute = self._append_attribute_to_path(
                        last_attribute,
                        attribute,
                    )

                attribute_type = self._get_attribute_type(
                    collection,
                    attribute,
                )

                aliased_attribute = self._get_aliased_path(
                    attribute,
                    alias_dict,
                )

                if collection != self._starting_collection:
                    aliased_attribute = self._append_attribute_to_path(
                        collection,
                        aliased_attribute,
                    )

                if attribute_type == str:
                    # print('Found froze: {}'.format(value))
                    if isinstance(value, frozenset):
                        # self.query_steps.extend(
                        #     unwind_stage,
                        # )
                        # for term in value:
                        term = " ".join(list(value))
                        term = self.substitute(term)
                        self._generate_string_match_stage(
                            aliased_attribute,
                            term,
                            collection=collection,
                            unwind_stage=unwind_stage,
                            # group=
                        )
                    else:
                        self._generate_string_match_stage(
                            aliased_attribute,
                            value,
                            collection=collection,
                            unwind_stage=unwind_stage,
                        )
                elif attribute_type in [int, float]:
                    self._generate_numeric_match_stage(
                        aliased_attribute,
                        value,
                        collection=collection,
                        unwind_stage=unwind_stage,
                    )
                elif attribute_type == list:
                    # print('Copying field2 ')
                    field_name, set_stage = self._copy_field(
                        attribute,
                    )

                    unwind_stage.append(
                        set_stage,
                    )

                    self.query_steps.extend(
                        unwind_stage,
                    )

                    alias_dict[attribute] = field_name

                    aliased_attribute = self._get_aliased_path(
                        attribute,
                        alias_dict,
                    )

                    self._generate_string_match_stage(
                        aliased_attribute,
                        value,
                        collection=collection,
                        unwind_stage=[
                            self.unwind(
                                aliased_attribute,
                            )
                        ],
                    )

    def substitute(self, term):
        d = {
            "depp johnny": "johnny depp",
            "jolie angelina": "angelina jolie",
            "washington denzel": "denzel washington",
            "ford harrison": "harrison ford",
            "eastwood clint": "clint eastwood",
            "smith will": "will smith",
            "hanks tom": "tom hanks",
            "freeman morgan": "morgan freeman",
            "wars star": "star wars",
            "gump forrest": "forrest gump",
            "depp johnny": "johnny depp",
            "jones indiana": "indiana jones",
            "bond james": "james bond",
            "lecter hannibal": "hannibal lecter",
            "bates norman": "normal bates",
            "vader darth": "darth vader",
            "lucas george": "george lucas",
            "ryan jack": "jack ryan",
            "curry stephen": "stephen curry",
            "angeles los": "los angeles",
            "koncak jon": "jon koncak",
            "bulls chicago": "chicago bulls",
            "james lebron": "lebron james",
            "anthony carmelo": "carmelo anthony",
            "jordan deandre": "deandre jordan",
            "cup world": "world_cup",
            "jackson percy": "percy jackson",
            "change climate": "climate change",
            "event apple": "apple event",
            "games hunger": "hunger games",
        }

        if term not in d:
            return term
        else:
            return d[term]

    def _check_for_list_in_parent_attributes(
        self, collection, attribute, parent_attributes, alias_dict
    ):
        for parent_attribute in parent_attributes:
            if parent_attribute == attribute:
                break

            if self._get_attribute_type(
                collection,
                parent_attribute,
            ):
                return True

        return False

    def _generate_string_match_stage(
        self, attribute, value, collection=None, unwind_stage=None
    ):
        if unwind_stage:
            self.query_steps.extend(
                unwind_stage,
            )

        self._match_string(
            attribute,
            value,
        )

        if unwind_stage:
            if not collection:
                raise TypeError("Missing parameter: collection")

            self._group(
                collection,
            )

    def _generate_numeric_match_stage(
        self, attribute, value, collection=None, unwind_stage=None
    ):
        if unwind_stage:
            self.query_steps.extend(
                unwind_stage,
            )

        self._match_number(
            attribute,
            value,
        )

        if unwind_stage:
            if not collection:
                raise TypeError("Missing parameter: collection")

            self._group(
                collection,
            )

    def _match_string(self, field_name, field_value):
        self.query_steps.append(
            {
                self.MATCH_STEP_FIELD: {
                    self.EXPRESSION_FIELD: self._regex_match(
                        field_name,
                        field_value,
                    )
                }
            }
        )

    def _match_number(self, field_name, field_value):
        new_field_value = None
        if "." in field_value:
            new_field_value = float(field_value)
        else:
            new_field_value = int(field_value)
        self.query_steps.append(
            {
                self.MATCH_STEP_FIELD: {
                    self.EXPRESSION_FIELD: {
                        "$eq": [
                            "".join(self._to_mongo_attribute_name(field_name)),
                            new_field_value,
                        ]
                    }
                }
            }
        )

    def _to_mongo_attribute_name(self, field_name):
        return "$" + field_name

    def _copy_field(self, field_name):
        dup_field_name = field_name + "_dup"
        return (
            dup_field_name,
            {
                self.SET_STEP_FIELD: {
                    dup_field_name: self._to_mongo_attribute_name(field_name)
                }
            },
        )

    def unwind(self, field_name, insert=False):
        if insert:
            self.query_steps.append(
                {
                    self.UNWIND_STEP_FIELD: self._to_mongo_attribute_name(
                        field_name
                    )
                }
            )

        return {
            self.UNWIND_STEP_FIELD: self._to_mongo_attribute_name(field_name)
        }

    def lookup(self, local_field, foreign_field, foreign_collection, alias):

        lookup_clause = {
            self.LOOKUP_STEP_FIELD: {
                "from": foreign_collection,
                "foreignField": foreign_field,
                "localField": local_field,
                "as": alias,
            }
        }

        self.query_steps.append(lookup_clause)
        self._last_added_step = self.LOOKUP_STEP_FIELD
        self._last_collection_lookup = foreign_collection

    def count_documents(self):
        count_clause = {self.COUNT_STEP_FIELD: "count"}
        self.query_steps.append(count_clause)

    def build(self):
        return (self._starting_collection, self.query_steps)



from math import log

"""
Weight Schemes recommended by Baeza-Yates

| Scheme | TF Weight                | IDF Weight        |
|--------|--------------------------|-------------------|
| 0      | raw frequency            | inverse frequency |
| 1      | log normalization        | unary             |
| 2      | log normalization        | inverse frequency |
| 3      | double normalization 0.5 | inverse frequency |

"""


def calculate_tf(weight_scheme, frequency, max_frequency=None):
    if weight_scheme == 0:
        # Raw Frequency
        return frequency
    if weight_scheme == 3:
        # Double Normalization 0.5
        return 0.5 + 0.5 * frequency / max_frequency
    if weight_scheme in [1, 2]:
        # Log Normalization
        return 1 + log(frequency)
    raise ValueError(f"Invalid weight_scheme({weight_scheme})")


def calculate_inverse_frequency(
    num_total_attributes, num_attributes_with_this_word
):
    return log(num_total_attributes / num_attributes_with_this_word)


def calculate_iaf(weight_scheme, inverse_frequency=None):
    if weight_scheme == 1:
        # Unary
        return 1
    if weight_scheme in [0, 2, 3]:
        # Inverse Frequency
        return inverse_frequency
    raise ValueError(f"Invalid weight_scheme({weight_scheme})")



import json
import logging
from copy import deepcopy
from typing import List

from polly.sources.utils.interfaces import SourceConfig

LOGGING_MAP = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "CRITICAL": logging.CRITICAL,
}


class ConfigHandler:
    __instance = None

    dataset = None
    queryset = None
    dataset_config = None
    queryset_config = None
    config_folder_directory = None
    connection = {}
    create_index = True
    results_directory = None
    plots_directory = None
    logging_mode = None
    default_settings = None
    sources_configs: List[SourceConfig] = []

    def __init__(self, config_directory, reset=False, **kwargs):
        if ConfigHandler.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            ConfigHandler.__instance = self

        self.config_folder_directory = config_directory
        self.configure(**kwargs)

    def configure(self, **kwargs):
        general_config_file = f"{self.config_folder_directory}config.json"

        config = self.load_config(general_config_file)

        self.sources_configs: List[SourceConfig] = []
        for source_config in config["sources"]:
            self.sources_configs.append(SourceConfig(**source_config))

        self.default_settings = deepcopy(config)
        self.dataset_config = config["dataset_config"]
        self.dataset_directory = self.dataset_config["dataset_directory"]
        self.querysets_directory = config["querysets_directory"]
        self.create_index = config["create_index"]
        self.plots_directory = config["plots_directory"]
        self.results_directory = config["results_directory"]
        self.attributes_filepath = self.dataset_config["attributes_filepath"]

        # if "handler_type" in config:
        #     self.handler_type = config["handler_type"]

        # if self.handler_type == DataSourceType.CSV.value:
        #     self.datasource_directory = self.dataset_config[
        #         "datasource_directory"
        #     ]

        # self.queryset_config = self.get_queryset_filepath(
        #     self.dataset,
        # )

        # self.update_config_paths(
        #     self.dataset,
        #     self.dataset_config,
        # )

        self.logging_mode = self.set_logging_level(
            config["logging_mode"],
        )

    def load_config(self, filepath):
        with open(filepath, "r") as f:
            config = json.load(f)

        return config

    def get_queryset_filepath(self, dataset):
        dataset_filename = dataset + ".json"
        if self.querysets_directory.endswith("/"):
            return "".join([self.querysets_directory, dataset_filename])

        return "/".join([self.querysets_directory, dataset_filename])

    def change_queryset(self, queryset):
        self.queryset = queryset

    def update_config_paths(self, value, config_paths):
        for configuration, path in config_paths.items():
            if configuration.endswith("filepath") or configuration.endswith(
                "directory"
            ):
                config_paths[configuration] = path.format(value)

    def set_logging_level(self, level):
        self.logging_mode = LOGGING_MAP[level]

    # def get_dataset_configs(self):
    #     subfolder = './datasets_config/'
    #     results = []
    #     for filepath in glob(f'{self.config_folder_directory}{subfolder}*.json'):
    #         with open(filepath,'r') as f:
    #             results.append( (json.load(f)['database'], filepath) )
    #     return results

    # def get_queryset_configs(self,dataset_config_filepath=None):
    #     subfolder = './querysets_config/'
    #     results = []
    #     for filepath in glob(f'{self.config_folder_directory}{subfolder}*.json'):
    #         with open(filepath,'r') as f:
    #             data = json.load(f)
    #             if dataset_config_filepath in (None,data['dataset_config_filepath']):
    #                 results.append( (data['queryset_name'], filepath) )
    #     return results



import json
from pathlib import Path
from typing import List

from polly.core.classes.query_match import QueryMatch
from polly.llm.chat import LLMChat


class LLMQMRanker:
    def __init__(
        self,
        keyword_query: str,
        database: str,
        query_matches: List[QueryMatch],
        llm_chat: LLMChat,
    ):
        self.keyword_query = keyword_query
        self.database = database
        self.query_matches = query_matches
        self.llm_chat = llm_chat

    def rank_query_matches(self):
        system_prompt = self._generate_system_prompt()
        system_message = self.llm_chat.build_system_message(system_prompt)

        user_prompt = self._generate_user_prompt()
        user_message = self.llm_chat.build_user_message(user_prompt)

        messages = [system_message, user_message]

        response = self.llm_chat.chat(messages, cast_response_to_json=True)

        query_matches: List[QueryMatch] = []
        for query_match_json in response:
            qm = QueryMatch.from_json_serializable(
                query_match_json, self.database
            )
            query_matches.append(qm)

        return query_matches

    def _generate_user_prompt(self):
        query_match_listing = ""

        for n, query_match in enumerate(self.query_matches):
            query_match_json = query_match.to_json_serializable()
            query_match_text = json.dumps(query_match_json)
            query_match_listing += f"{n + 1}. {query_match_text}\n"

        user_prompt = f"""
            Keyword Query: {self.keyword_query}

            Task: Consider you mapped each keyword from the query to the most relevant table(s) and column(s) in the {self.database} database, based on the description of the table(s) and column(s). These mappings were combined to generate what we call "Query Matches", that are minimal set covers for the keywords. These covers have 2 properties: totality, in which the cover contains all keywords from the query, and minimality, in which we cannot remove any mapping such that the query match is not total.

            Instructions:
            1. Return an ordered list from the highest to lowest probable correct query match provided as input.
            2. Prioritize mappings keywords to the most relevant tables or attributes in the schema, and relate them in the Query Match using join logic. Focus on minimal sets that cover only the relationship, avoiding non-explicit attributes in the Query Match.
            3. Mappings such as "self" should be prioritized over non-explicit mappings.
            4. Schema mappings that do not present a direct and meaningful relationship with the keyword query should be considered less relevant.
            5. Return the list in plain text using the JSON schema below. Do not output markdown.

            Final output only according to the example json schema below in plain text:
            [
                [{{"table": "tableName", "schema_filter": [], "value_filter": [{{"attribute": "attributeName", "keywords": ["hungary"]}}]}}, {{"table": "tableName", "schema_filter": [], "value_filter": [{{"attribute": "attributeName", "keywords": ["slovakia"]}}]}}],
                [{{"table": "tableName", "schema_filter": [], "value_filter": [{{"attribute": "attributeName", "keywords": ["hungary"]}}]}}, {{"table": "tableName", "schema_filter": [], "value_filter": [{{"attribute": "attributeName", "keywords": ["slovakia"]}}]}}]
            ]

            Input:
            {query_match_listing}
        """

        return user_prompt

    def _generate_system_prompt(self):
        system_prompt = f"Below are the tables from the {self.database} database with their column names and descriptions in JSON format: \n"

        schema_description_filepath = f"tmp/{self.database}.json"

        # print(f"Opening {schema_description_filepath}")

        if not Path(schema_description_filepath).exists():
            raise Exception(
                f"Schema description file not found: {schema_description_filepath}"
            )

        with open(schema_description_filepath, "r") as file:
            system_prompt += file.read()

        return system_prompt



from typing import List

from pydantic import BaseModel

from polly.sources.enum import DataSourceType


class ConnectionConfig(BaseModel):
    host: str
    port: int
    user: str
    password: str
    database: str = ""
    collections: List[str] = []


class SourceConfig(BaseModel):
    name: str
    type: DataSourceType
    connection_config: ConnectionConfig



import json
import re
import string


class KeywordMatch:

    def __init__(
        self,
        table: str,
        database: str,
        value_filter={},
        schema_filter={},
    ):
        self.__slots__ = ["table", "schema_filter", "value_filter"]
        self.table = table
        self.database = database
        self.schema_filter = frozenset(
            {
                (attribute, frozenset(keywords))
                for attribute, keywords in schema_filter.items()
            }
        )
        self.value_filter = frozenset(
            {
                (attribute, frozenset(keywords))
                for attribute, keywords in value_filter.items()
            }
        )

    def is_free(self):
        return len(self.schema_filter) == 0 and len(self.value_filter) == 0

    def add_schema_filter(self, attribute, keywords):
        self.schema_filter = frozenset(
            self.schema_filter.union({(attribute, frozenset(keywords))})
        )

    def add_value_filter(self, attribute, keywords):
        self.value_filter = frozenset(
            self.value_filter.union({(attribute, frozenset(keywords))})
        )

    def has_value_mmapings(self):
        return len(self.value_filter) > 0

    def has_schema_mmapings(self):
        return len(self.schema_filter) > 0

    def schema_mappings(self):
        for attribute, keywords in self.schema_filter:
            yield (self.table, attribute, keywords)

    def value_mappings(self):
        for attribute, keywords in self.value_filter:
            yield (self.table, attribute, keywords)

    def has_value_mappings(self):
        return self.value_filter != frozenset({})

    def mappings(self):
        for attribute, keywords in self.schema_filter:
            yield ("s", self.table, attribute, keywords)
        for attribute, keywords in self.value_filter:
            yield ("v", self.table, attribute, keywords)

    def keywords(self, schema_only=False):
        for _, keywords in self.schema_filter:
            yield from keywords
        if not schema_only:
            for _, keywords in self.value_filter:
                yield from keywords

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        def str_filter(filter_type, fltr):
            if len(fltr) == 0:
                return ""

            return ".{}({})".format(
                filter_type,
                ",".join(
                    [
                        "{}{{{}}}".format(attribute, ",".join(keywords))
                        for attribute, keywords in fltr
                    ]
                ),
            )

        return "{}:{}{}{}".format(
            self.database.upper(),
            self.table.upper(),
            str_filter("s", self.schema_filter),
            str_filter("v", self.value_filter),
        )

    def __eq__(self, other):
        return (
            isinstance(other, KeywordMatch)
            and self.table == other.table
            # and set(self.keywords(schema_only=True)) == set(other.keywords(schema_only=True))
            and self.schema_filter == other.schema_filter
            and self.value_filter == other.value_filter
        )

    def has_default_mapping(self):
        schema_items = [
            attr for (attr, keywords) in self.schema_filter if attr == "*"
        ]
        value_items = [
            attr for (attr, keywords) in self.value_filter if attr == "*"
        ]

        return len(value_items) > 0 or len(schema_items) > 0

    def replace_default_mapping(self, attribute):
        schema_items = [
            attr for (attr, keywords) in self.schema_filter if attr == "*"
        ]
        value_items = [
            attr for (attr, keywords) in self.value_filter if attr == "*"
        ]

        if len(value_items) > 0:
            new_value_set = set(
                {
                    (attr, keywords)
                    for attr, keywords in self.value_filter
                    if attr != "*"
                }
            )
            value = [
                keywords
                for (attr, keywords) in self.value_filter
                if attr == "*"
            ][0]
            new_value_set.add((attribute, value))
            self.value_filter = frozenset(new_value_set)

        if len(schema_items) > 0:
            new_schema_set = set(
                {
                    (attr, keywords)
                    for attr, keywords in self.schema_filter
                    if attr != "*"
                }
            )
            value = [
                keywords
                for (attr, keywords) in self.schema_filter
                if attr == "*"
            ][0]
            new_schema_set.add((attribute, value))
            self.schema_filter = frozenset(new_schema_set)

    def has_same_attribute(self, other):
        all_equal_attributes = False
        attrib = set([attrib for (attrib, keywords) in self.value_filter])
        other_attrib = set(
            [attrib for (attrib, keywords) in other.value_filter]
        )

        if attrib != other_attrib:
            return False

        attrib = set([attrib for (attrib, keywords) in self.schema_filter])
        other_attrib = set(
            [attrib for (attrib, keywords) in other.schema_filter]
        )

        if attrib == other_attrib:
            all_equal_attributes = True

        return all_equal_attributes

    def __hash__(self):
        # return hash( (self.table,frozenset(self.keywords(schema_only=True)),self.value_filter) )
        return hash((self.table, self.schema_filter, self.value_filter))

    def to_json_serializable(self):

        def filter_object(fltr):
            return [
                {"attribute": attribute, "keywords": list(keywords)}
                for attribute, keywords in fltr
            ]

        return {
            "table": self.table,
            "database": self.database,
            "schema_filter": filter_object(self.schema_filter),
            "value_filter": filter_object(self.value_filter),
        }

    def to_json(self):
        return json.dumps(self.to_json_serializable())

    @staticmethod
    def from_str(str_km):
        re_km = re.compile(r"([A-Z,_,1-9]+)(.*)")
        re_filters = re.compile(r"\.([vs])\(([^\)]*)\)")
        re_predicates = re.compile(r"([\w\*]*)\{([^\}]*)\}\,?")
        re_keywords = re.compile(r"(\w+)\,?")

        m_km = re_km.match(str_km)
        table = m_km.group(1).lower()
        schema_filter = {}
        value_filter = {}
        for filter_type, str_predicates in re_filters.findall(m_km.group(2)):

            if filter_type == "v":
                predicates = value_filter
            else:
                predicates = schema_filter
            for attribute, str_keywords in re_predicates.findall(
                str_predicates
            ):
                predicates[attribute] = {
                    key for key in re_keywords.findall(str_keywords)
                }
        return KeywordMatch(
            table,
            value_filter=value_filter,
            schema_filter=schema_filter,
        )

    @staticmethod
    def from_json_serializable(json_serializable, database: str):

        def filter_dict(filter_obj):
            return {
                predicate["attribute"]: [
                    keyword for keyword in predicate["keywords"]
                ]
                for predicate in filter_obj
            }

        return KeywordMatch(
            json_serializable["table"],
            database,
            value_filter=filter_dict(json_serializable["value_filter"]),
            schema_filter=filter_dict(json_serializable["schema_filter"]),
        )

    @staticmethod
    def from_json(str_json):
        return KeywordMatch.from_json_serializable(json.loads(str_json))



import psycopg2
from psycopg2.sql import SQL, Identifier

from polly.utils import Tokenizer


class PostgresIter:
    def __init__(self, config, database_table_columns, **kwargs):
        self.limit_per_table = kwargs.get("limit_per_table", None)
        self.tokenizer = kwargs.get("tokenizer", Tokenizer())
        self.config = config
        self.database_table_columns = database_table_columns

        self.table_hash = self._get_indexable_schema_elements()

    def _schema_element_validator(self, table, column):
        return True

    def _get_indexable_schema_elements(self):
        with psycopg2.connect(**self.config.connection) as conn:
            with conn.cursor() as cur:
                GET_TABLE_AND_COLUMNS_WITHOUT_FOREIGN_KEYS_SQL = """
                    SELECT
                        c.table_name,
                        c.column_name
                    FROM
                        information_schema.table_constraints AS tc
                        JOIN information_schema.key_column_usage AS kcu
                          ON tc.constraint_name = kcu.constraint_name
                          AND tc.table_schema = kcu.table_schema
                          AND tc.constraint_type = 'FOREIGN KEY'
                        RIGHT JOIN information_schema.columns AS c
                          ON c.table_name=tc.table_name
                          AND c.column_name = kcu.column_name
                          AND c.table_schema = kcu.table_schema
                    WHERE
                        c.table_schema='public'
                        AND tc.constraint_name IS NULL;
                """
                cur.execute(GET_TABLE_AND_COLUMNS_WITHOUT_FOREIGN_KEYS_SQL)
                table_hash = {}
                for table, column in cur.fetchall():
                    if (table, column) not in self.database_table_columns:
                        print(f"SKIPPED {(table,column)}")
                        continue
                    table_hash.setdefault(table, []).append(column)
                return table_hash

    def __iter__(self):
        for table, columns in self.table_hash.items():

            indexable_columns = [
                col
                for col in columns
                if self._schema_element_validator(table, col)
            ]

            if len(indexable_columns) == 0:
                continue

            print(
                "\nINDEXING {}({})".format(table, ", ".join(indexable_columns))
            )

            with psycopg2.connect(**self.config.connection) as conn:
                with conn.cursor() as cur:
                    """
                    NOTE: Table and columns can't be directly passed as parameters.
                    Thus, the psycopg2.sql.SQL command with psycopg2.sql.Identifiers is used
                    """

                    if self.limit_per_table is not None:
                        text = f"SELECT ctid, {{}} FROM {{}} LIMIT {self.limit_per_table};"
                    else:
                        text = "SELECT ctid, {} FROM {};"

                    cur.execute(
                        SQL(text).format(
                            SQL(", ").join(
                                Identifier(col) for col in indexable_columns
                            ),
                            Identifier(table),
                        )
                    )

                    arraysize = 100000
                    data = cur.fetchmany(arraysize)
                    while len(data) != 0:
                        for row in data:
                            ctid = row[0]
                            for col in range(1, len(row)):
                                column = cur.description[col][0]
                                text = str(row[col])
                                tokens = self.tokenizer.tokenize(text)

                                for word in tokens:
                                    # print(tokens)
                                    # if len([ch for ch in word if ch in string.punctuation]) != 0 and not is_url:
                                    #     print("Tokenizer not working {}".format(word, row[col]))
                                    #     sys.exit()
                                    yield table, ctid, column, word

                        data = cur.fetchmany(arraysize)



from copy import deepcopy
from queue import deque
from typing import Dict, List

from sentence_transformers import SentenceTransformer

from polly.core.classes.candidate_network import CandidateNetwork
from polly.core.classes.keyword_match import KeywordMatch
from polly.core.classes.query_match import QueryMatch
from polly.sources.handler import SourceHandler
from polly.utils.schema_graph import SchemaGraph


class CandidateNetworkHandler:
    def __init__(self, source_handler: SourceHandler):
        self.source_handler = source_handler
        self.bert_model = SentenceTransformer("all-mpnet-base-v2")

    def generate_cns(
        self,
        schema_index: Dict[str, Dict[str, str]],
        schema_graph: SchemaGraph,
        ranked_query_matches: List[QueryMatch],
        keywords: List[str],
        keyword_query: str,
        **kwargs
    ):
        topk_cns = kwargs.get("topk_cns", 20)
        cjn_ranking_method = kwargs.get("cjn_ranking_method", 0)

        returned_cns = []
        num_cns_available = topk_cns

        for query_match in ranked_query_matches:
            if topk_cns != -1 and num_cns_available <= 0:
                break

            cns_per_cur_qm = self.generate_cns_per_qm(
                schema_index, schema_graph, query_match, keywords, **kwargs
            )

            for candidate_network in cns_per_cur_qm:
                if num_cns_available <= 0:
                    break

                returned_cns.append(candidate_network)

                # candidate_network.calculate_score(
                #     query_match,
                #     keyword_query,
                #     schema_graph,
                #     self.database_handler,
                #     self.bert_model,
                #     **kwargs
                # )

                num_cns_available -= 1

        return returned_cns

    def generate_cns_per_qm(
        self, schema_index, schema_graph, query_match, keywords, **kwargs
    ):
        max_cn_size = kwargs.get("max_cn_size", 5)
        topk_cns_per_qm = kwargs.get("topk_cns_per_qm", 1)

        instance_based_pruning = kwargs.get("instance_based_pruning", False)
        max_database_accesses = kwargs.get("max_database_accesses", 7)

        # directed_neighbor_sorting_function = kwargs.get(
        #     "directed_neighbor_sorting_function",
        #     self.factory_sum_norm_attributes(schema_index, weight_scheme),
        # )
        schema_prunning = kwargs.get("schema_prunning", True)
        desired_cn = kwargs.get("desired_cn", None)

        returned_cns = []
        empty_cns = []
        pruned_cns = set()

        def meet_pruning_conditions(jnkm):
            return not (
                # corretude rules
                jnkm.is_sound(schema_graph)
                and
                # checking whether it was already generated
                jnkm not in F
                and jnkm not in pruned_cns
                and jnkm not in returned_cns
                and jnkm not in empty_cns
                and
                # max cn size pruning
                len(jnkm) <= max_cn_size
                and (
                    not schema_prunning
                    or (
                        # max number of leaves pruning
                        jnkm.num_leaves() <= len(query_match)
                        and
                        # max number of keyword-free matches pruning
                        jnkm.num_free_keyword_matches() + len(query_match)
                        <= max_cn_size
                    )
                )
            )

        # JNKM stands for Joining Network of Keyword Matches
        cur_jnkm = CandidateNetwork()
        cur_jnkm.add_keyword_match(query_match.get_km_from_keyword(keywords[0]))

        if len(query_match) == 1:
            returned_cns.append(cur_jnkm)
            return returned_cns

        table_hash = {}
        for keyword_match in query_match:
            table_hash.setdefault(
                keyword_match.table,
                {
                    KeywordMatch(
                        keyword_match.table, self.source_handler.database
                    )
                },
            ).add(keyword_match)

        F = deque()
        F.append(cur_jnkm)

        while F:
            cur_jnkm = F.popleft()

            for vertex_u in reversed(list(cur_jnkm.vertices())):
                keyword_match, _ = vertex_u

                directed_neighbors = schema_graph.directed_neighbors(
                    keyword_match.table
                )

                for direction, adj_table in directed_neighbors:

                    table_hash.setdefault(
                        adj_table,
                        {KeywordMatch(adj_table, self.source_handler.database)},
                    )

                    for adj_keyword_match in table_hash[adj_table]:

                        if (
                            not (adj_keyword_match.table == keyword_match.table)
                            and adj_keyword_match
                            not in cur_jnkm.keyword_matches()
                            or adj_keyword_match.is_free()
                        ):

                            next_jnkm = deepcopy(cur_jnkm)
                            next_jnkm.add_adjacent_keyword_match(
                                vertex_u,
                                adj_keyword_match,
                                edge_direction=direction,
                            )

                            if not meet_pruning_conditions(next_jnkm):
                                # print(f'next_jnkm:\n{next_jnkm}')
                                if next_jnkm.is_total(query_match):
                                    if (
                                        not next_jnkm.contains_keyword_free_match_leaf()
                                    ):
                                        if (
                                            not instance_based_pruning
                                            or not self.is_cn_empty(
                                                schema_graph,
                                                schema_index,
                                                next_jnkm,
                                            )
                                        ):
                                            returned_cns.append(next_jnkm)
                                        else:
                                            empty_cns.append(next_jnkm)

                                        if (
                                            desired_cn is not None
                                            and next_jnkm == desired_cn
                                        ):
                                            return returned_cns
                                    else:
                                        # reduce JNKM
                                        # print(f'next_jnkm(pruned): non minimal\n{next_jnkm}')
                                        pruned_cns.add(next_jnkm)
                                        continue

                                    if len(returned_cns) >= topk_cns_per_qm or (
                                        instance_based_pruning
                                        and len(returned_cns) + len(empty_cns)
                                        >= max_database_accesses
                                    ):
                                        return returned_cns

                                elif len(next_jnkm) < max_cn_size:
                                    F.append(next_jnkm)
                                    # print(f'next_jnkm:\n{next_jnkm}')
                                else:
                                    pruned_cns.add(next_jnkm)
                            else:
                                pruned_cns.add(next_jnkm)

        return returned_cns

    def factory_sum_norm_attributes(self, schema_index, weight_scheme):

        def sum_norm_attributes(directed_neighbor):
            _, adj_table = directed_neighbor
            score1 = 0
            if adj_table in schema_index:
                score1 = sum(
                    schema_index[adj_table][attribute]["norm"][weight_scheme]
                    for attribute in schema_index[adj_table]
                )
            score2 = adj_table
            return (-score1, score2)

        return sum_norm_attributes

    def is_cn_empty(self, schema_graph, schema_index, candidate_network):
        sql = candidate_network.get_sql_from_cn(
            schema_graph, schema_index[self.source_handler.database]
        )
        return self.database_handler.exist_results(sql) == False



from IPython.display import display
from pandas import DataFrame, MultiIndex


class EvaluationResult:

    def __init__(self, data):
        self.data = data

    def metrics(self, show=True):
        data = {
            "QMs": self.data["evaluation"]["query_matches"],
            "CJNs": self.data["evaluation"]["candidate_networks"],
        }
        df = DataFrame.from_dict(data, orient="index")
        del df["relevant_positions"]
        df.columns = df.columns.str.upper()
        df = df.round(2)
        if show:
            display(df)
        return df

    def relevant_positions(self):
        data = {
            "Keyword Query": [
                item["keyword_query"] for item in self.data["results"]
            ],
            "Relevant QM": self.data["evaluation"]["query_matches"][
                "relevant_positions"
            ],
            "Relevant CJN": self.data["evaluation"]["candidate_networks"][
                "relevant_positions"
            ],
        }
        df = DataFrame.from_dict(data)
        df.index = df.index + 1
        df.columns = MultiIndex(
            levels=[["", "Relevant Position"], ["Keyword Query", "QM", "CJN"]],
            codes=[[0, 1, 1], [0, 1, 2]],
        )
        display(df)
        return df



import datetime
from copy import deepcopy
from typing import List, Tuple

from bson.int64 import Int64


class DocumentTraverser:

    def __init__(self):
        self.traversed_attributes = {}
        self.document_attributes = {}
        self.document_attributes_values = {}
        self.indexable_content = []

    def _traverse(self, attribute_path, document_path, value, value_type):
        if value_type == list:
            self.document_attributes[document_path] = value_type
            for number, element in enumerate(value):
                updated_path = "".join(
                    [
                        attribute_path,
                        "[",
                        str(number + 1),
                        "]",
                    ]
                )
                self._add_to_document_attributes(document_path, type(element))
                self._traverse(
                    updated_path, document_path, element, type(element)
                )
        elif value_type == dict:
            if document_path not in self.document_attributes:
                self.document_attributes[document_path] = value_type
            for attribute, attribute_value in value.items():
                updated_path = self._increment_document_path(
                    attribute_path, attribute
                )
                updated_document_path = self._increment_document_path(
                    document_path, attribute
                )
                self._add_to_document_attributes(
                    updated_document_path, type(attribute_value)
                )
                self._traverse(
                    updated_path,
                    updated_document_path,
                    attribute_value,
                    type(attribute_value),
                )
        elif value_type in [str, int, float, Int64]:
            self._add_to_document_attributes(document_path, value_type)
            self.traversed_attributes[attribute_path] = value
            self.indexable_content.append(
                (document_path, value),
            )
            self.document_attributes_values[document_path] = value
        elif value_type == datetime.datetime:
            self._add_to_document_attributes(document_path, value_type)
            self.traversed_attributes[attribute_path] = value.strftime(
                "%d/%m/%Y %H:%M:%S"
            )
            self.document_attributes_values[document_path] = value

    def _add_to_document_attributes(self, document_path, value_type):
        if document_path not in self.document_attributes:
            self.document_attributes[document_path] = value_type

    def _increment_document_path(self, current_path, new_attribute):
        return ".".join([current_path, new_attribute])

    def cleanup(self):
        self.traversed_attributes = {}
        # TODO: evaluate if we can mantain this dict while iterating through all documents
        # self.document_attributes = {}
        self.indexable_content = []

    def traverse(self, document):
        for attribute, value in document.items():
            # print(attribute, type(value))
            self._traverse(attribute, attribute, value, type(value))

    def get_traversed_attributes(self):
        return self.traversed_attributes

    def get_document_attributes(self):
        return self.document_attributes

    def get_document_attributes_values(self):
        return self.document_attributes_values

    def convert_document_attributes_values_to_list_of_tuples(self):
        l: List[Tuple[str, str]] = []
        for key, value in self.document_attributes_values.items():
            l.append((key, value))

        return l

    def get_document_attributes_without_top_level_attributes(self):
        filtered_document_attributes = deepcopy(self.document_attributes)
        attributes_for_removal = []
        for attribute in filtered_document_attributes:
            if "." not in attribute and filtered_document_attributes[
                attribute
            ] in [type(list()), type(dict())]:
                # print('Removing attribute from collection structure')
                attributes_for_removal.append(attribute)
            else:
                if filtered_document_attributes[attribute] in [
                    type(list()),
                    type(dict()),
                ]:
                    attributes_for_removal.append(attribute)
        for attribute in attributes_for_removal:
            del filtered_document_attributes[attribute]

        return filtered_document_attributes

    def get_indexable_content(self):
        return self.indexable_content


if __name__ == "__main__":
    document = {
        "_id": {"$oid": "52f587f0b42a75c4264022c4"},
        "teams": [
            {
                "name": "Washington Wizards",
                "abbreviation": "WAS",
                "score": 78,
                "home": True,
                "won": 0,
                "results": {
                    "ast": 13,
                    "blk": 2,
                    "drb": 30,
                    "fg": 27,
                    "fg3": 4,
                    "fg3_pct": ".211",
                    "fg3a": 19,
                    "fg_pct": ".375",
                    "fga": 72,
                    "ft": 20,
                    "ft_pct": ".714",
                    "fta": 28,
                    "mp": 240,
                    "orb": 9,
                    "pf": 18,
                    "plus_minus": "",
                    "pts": 78,
                    "stl": 8,
                    "tov": 14,
                    "trb": 39,
                },
                "players": [
                    {
                        "ast": 1,
                        "blk": 0,
                        "drb": 6,
                        "fg": 4,
                        "fg3": 0,
                        "fg3_pct": "",
                        "fg3a": 0,
                        "fg_pct": ".500",
                        "fga": 8,
                        "ft": 5,
                        "ft_pct": ".833",
                        "fta": 6,
                        "mp": "35:00",
                        "orb": 5,
                        "pf": 4,
                        "player": "Jared Jeffries",
                        "plus_minus": -9,
                        "pts": 13,
                        "stl": 0,
                        "tov": 2,
                        "trb": 11,
                    },
                    {
                        "ast": 0,
                        "blk": 0,
                        "drb": 2,
                        "fg": 5,
                        "fg3": 1,
                        "fg3_pct": ".250",
                        "fg3a": 4,
                        "fg_pct": ".500",
                        "fga": 10,
                        "ft": 4,
                        "ft_pct": ".667",
                        "fta": 6,
                        "mp": "33:00",
                        "orb": 1,
                        "pf": 1,
                        "player": "Jarvis Hayes",
                        "plus_minus": "+2",
                        "pts": 15,
                        "stl": 1,
                        "tov": 0,
                        "trb": 3,
                    },
                    {
                        "ast": 3,
                        "blk": 1,
                        "drb": 2,
                        "fg": 1,
                        "fg3": 1,
                        "fg3_pct": ".200",
                        "fg3a": 5,
                        "fg_pct": ".111",
                        "fga": 9,
                        "ft": 0,
                        "ft_pct": "",
                        "fta": 0,
                        "mp": "30:00",
                        "orb": 0,
                        "pf": 0,
                        "player": "Steve Blake",
                        "plus_minus": "+4",
                        "pts": 3,
                        "stl": 1,
                        "tov": 3,
                        "trb": 2,
                    },
                    {
                        "ast": 2,
                        "blk": 1,
                        "drb": 5,
                        "fg": 3,
                        "fg3": 0,
                        "fg3_pct": ".000",
                        "fg3a": 1,
                        "fg_pct": ".300",
                        "fga": 10,
                        "ft": 2,
                        "ft_pct": "1.000",
                        "fta": 2,
                        "mp": "26:00",
                        "orb": 0,
                        "pf": 2,
                        "player": "Larry Hughes",
                        "plus_minus": "+7",
                        "pts": 8,
                        "stl": 0,
                        "tov": 0,
                        "trb": 5,
                    },
                    {
                        "ast": 2,
                        "blk": 0,
                        "drb": 4,
                        "fg": 4,
                        "fg3": 0,
                        "fg3_pct": "",
                        "fg3a": 0,
                        "fg_pct": ".800",
                        "fga": 5,
                        "ft": 0,
                        "ft_pct": "",
                        "fta": 0,
                        "mp": "18:00",
                        "orb": 1,
                        "pf": 2,
                        "player": "Brendan Haywood",
                        "plus_minus": "+2",
                        "pts": 8,
                        "stl": 0,
                        "tov": 0,
                        "trb": 5,
                    },
                    {
                        "ast": 0,
                        "blk": 0,
                        "drb": 5,
                        "fg": 6,
                        "fg3": 0,
                        "fg3_pct": "",
                        "fg3a": 0,
                        "fg_pct": ".667",
                        "fga": 9,
                        "ft": 4,
                        "ft_pct": ".667",
                        "fta": 6,
                        "mp": "30:00",
                        "orb": 1,
                        "pf": 3,
                        "player": "Etan Thomas",
                        "plus_minus": -18,
                        "pts": 16,
                        "stl": 3,
                        "tov": 2,
                        "trb": 6,
                    },
                    {
                        "ast": 1,
                        "blk": 0,
                        "drb": 1,
                        "fg": 3,
                        "fg3": 1,
                        "fg3_pct": ".200",
                        "fg3a": 5,
                        "fg_pct": ".273",
                        "fga": 11,
                        "ft": 1,
                        "ft_pct": ".500",
                        "fta": 2,
                        "mp": "21:00",
                        "orb": 1,
                        "pf": 2,
                        "player": "Juan Dixon",
                        "plus_minus": -18,
                        "pts": 8,
                        "stl": 1,
                        "tov": 1,
                        "trb": 2,
                    },
                    {
                        "ast": 2,
                        "blk": 0,
                        "drb": 2,
                        "fg": 0,
                        "fg3": 0,
                        "fg3_pct": ".000",
                        "fg3a": 3,
                        "fg_pct": ".000",
                        "fga": 5,
                        "ft": 2,
                        "ft_pct": ".500",
                        "fta": 4,
                        "mp": "20:00",
                        "orb": 0,
                        "pf": 2,
                        "player": "Gilbert Arenas",
                        "plus_minus": -25,
                        "pts": 2,
                        "stl": 1,
                        "tov": 4,
                        "trb": 2,
                    },
                    {
                        "ast": 0,
                        "blk": 0,
                        "drb": 2,
                        "fg": 1,
                        "fg3": 1,
                        "fg3_pct": "1.000",
                        "fg3a": 1,
                        "fg_pct": ".250",
                        "fga": 4,
                        "ft": 0,
                        "ft_pct": "",
                        "fta": 0,
                        "mp": "15:00",
                        "orb": 0,
                        "pf": 0,
                        "player": "Mitchell Butler",
                        "plus_minus": -18,
                        "pts": 3,
                        "stl": 1,
                        "tov": 1,
                        "trb": 2,
                    },
                    {
                        "ast": 2,
                        "blk": 0,
                        "drb": 1,
                        "fg": 0,
                        "fg3": 0,
                        "fg3_pct": "",
                        "fg3a": 0,
                        "fg_pct": ".000",
                        "fga": 1,
                        "ft": 2,
                        "ft_pct": "1.000",
                        "fta": 2,
                        "mp": "12:00",
                        "orb": 0,
                        "pf": 2,
                        "player": "Lonny Baxter",
                        "plus_minus": -7,
                        "pts": 2,
                        "stl": 0,
                        "tov": 1,
                        "trb": 1,
                    },
                ],
                "city": "Washington",
            }
        ],
        "date": {"$date": "2004-04-14T04:00:00.000+0000"},
    }
    from pprint import pprint as pp

    d = DocumentTraverser()
    d.traverse(document)
    pp(d.convert_document_attributes_values_to_dict())



import json
from collections import OrderedDict
from copy import deepcopy
from os import makedirs
from os.path import dirname

from polly.core.classes.candidate_network import CandidateNetwork
from polly.core.classes.query_match import QueryMatch
from polly.evaluation.result import EvaluationResult
from polly.utils.path import last_path, next_path


class EvaluationHandler:

    def __init__(self, config):
        self.config = config

    def load_golden_standards(self):
        with open(self.config.queryset_filepath, mode="r") as f:
            data = json.load(f)

        golden_standards = OrderedDict()
        for item in data:
            if "query_matches" in item:
                item["query_matches"] = [
                    QueryMatch.from_json_serializable(json_serializable_qm)
                    for json_serializable_qm in item["query_matches"]
                ]

            if "candidate_networks" in item:
                item["candidate_networks"] = [
                    CandidateNetwork.from_json_serializable(
                        json_serializable_cn
                    )
                    for json_serializable_cn in item["candidate_networks"]
                ]

            golden_standards[item["keyword_query"]] = item

        self.golden_standards = golden_standards

    def load_results_from_file(self, **kwargs):
        approach = kwargs.get("approach", "standard")
        results_filename = kwargs.get(
            "results_filename",
            last_path(
                f"{self.config.results_directory}{self.config.queryset_name}-{approach}-%03d.json"
            ),
        )

        with open(results_filename, mode="r") as f:
            data = json.load(f)
        return data

    def evaluate_results(self, results, **kwargs):
        """
        results_filename is declared here for sake of readability. But the
        default value is only set later to get a timestamp more accurate.
        """
        results_filename = kwargs.get("results_filename", None)
        approach = kwargs.get("approach", "standard")
        skip_ranking_evaluation = kwargs.get("skip_ranking_evaluation", False)
        skip_cn_ranking_evaluation = kwargs.get(
            "skip_cn_ranking_evaluation", False
        )
        write_evaluation_only = kwargs.get("write_evaluation_only", False)

        in_place_results = kwargs.get("in_place_results", False)
        use_result_class = kwargs.get("use_result_class", True)

        if not in_place_results:
            results = deepcopy(results)
        results["results"] = [
            item.data if isinstance(item, LatheResult) else item
            for item in results["results"]
        ]

        if not skip_ranking_evaluation:
            self.evaluate_query_matches(results)
            if not skip_cn_ranking_evaluation:
                self.evaluate_candidate_networks(results)
        self.evaluate_performance(results)
        self.evaluate_num_keyword_matches(results)
        self.evaluate_num_query_matches(results)
        self.evaluate_num_candidate_networks(results)

        if write_evaluation_only:
            del results["results"]

        if results_filename is None:
            makedirs(self.config.results_directory, exist_ok=True)
            results_filename = next_path(
                f"{self.config.results_directory}{self.config.queryset_name}-{approach}-%03d.json"
            )
        else:
            makedirs(dirname(results_filename), exist_ok=True)

        with open(results_filename, mode="w") as f:
            logger.info(f"Writing evaluated results in {results_filename}")
            json.dump(results, f, indent=4)

        if use_result_class:
            return EvaluationResult(results)

        return results

    def evaluate_query_matches(self, results, **kwargs):
        max_k = kwargs.get("max_k", 10)

        results.setdefault("evaluation", {})
        results["evaluation"]["query_matches"] = {}

        relevant_positions = []
        for item in results["results"]:
            if isinstance(item, LatheResult):
                item = item.data
            if "query_matches" in item:
                golden_qm = self.golden_standards[item["keyword_query"]][
                    "query_matches"
                ][0]

                query_matches = [
                    QueryMatch.from_json_serializable(json_serializable_qm)
                    for json_serializable_qm in item["query_matches"]
                ]

                relevant_position = self.get_relevant_position(
                    query_matches, golden_qm
                )

                relevant_positions.append(relevant_position)

        results["evaluation"]["query_matches"]["mrr"] = (
            self.get_mean_reciprocal_rank(relevant_positions)
        )

        precision_at_k = {
            f"p@{k}": self.get_precision_at_k(k, relevant_positions)
            for k in range(1, max_k + 1)
        }
        results["evaluation"]["query_matches"].update(precision_at_k)

        results["evaluation"]["query_matches"][
            "relevant_positions"
        ] = relevant_positions

        # print('QM Evaluation {}'.format(results['evaluation']['query_matches']))

    def evaluate_candidate_networks(self, results, **kwargs):
        max_k = kwargs.get("max_k", 10)

        results.setdefault("evaluation", {})
        results["evaluation"]["candidate_networks"] = {}

        relevant_positions = []
        for item in results["results"]:
            if isinstance(item, LatheResult):
                item = item.data
            if "candidate_networks" in item:
                golden_cn = self.golden_standards[item["keyword_query"]][
                    "candidate_networks"
                ][0]
                # print(f'Golden CN:\n{golden_cn}')

                candidate_networks = [
                    CandidateNetwork.from_json_serializable(
                        json_serializable_cn
                    )
                    for json_serializable_cn in item["candidate_networks"]
                ]

                # for i,candidate_network in enumerate(candidate_networks):
                #     print(f'{i+1} CN:\n{candidate_network}')

                relevant_position = self.get_relevant_position(
                    candidate_networks, golden_cn
                )

                relevant_positions.append(relevant_position)

        results["evaluation"]["candidate_networks"]["mrr"] = (
            self.get_mean_reciprocal_rank(relevant_positions)
        )

        precision_at_k = {
            f"p@{k}": self.get_precision_at_k(k, relevant_positions)
            for k in range(1, max_k + 1)
        }
        results["evaluation"]["candidate_networks"].update(precision_at_k)

        results["evaluation"]["candidate_networks"][
            "relevant_positions"
        ] = relevant_positions

        # print('CN Evaluation {}'.format(results['evaluation']['candidate_networks']))

    def evaluate_performance(self, results, **kwargs):
        results.setdefault("evaluation", {})
        results["evaluation"]["performance"] = {}

        for item in results["results"]:
            if isinstance(item, LatheResult):
                item = item.data
            if "elapsed_time" in item:
                for phase in item["elapsed_time"]:
                    results["evaluation"]["performance"].setdefault(
                        phase, []
                    ).append(item["elapsed_time"][phase])

    def evaluate_num_keyword_matches(self, results, **kwargs):
        results.setdefault("evaluation", {})
        results["evaluation"]["num_keyword_matches"] = []

        for item in results["results"]:
            if isinstance(item, LatheResult):
                item = item.data
            if "num_keyword_matches" in item:
                results["evaluation"]["num_keyword_matches"].append(
                    item["num_keyword_matches"]
                )

    def evaluate_num_query_matches(self, results, **kwargs):
        results.setdefault("evaluation", {})
        results["evaluation"]["num_query_matches"] = []

        for item in results["results"]:
            if isinstance(item, LatheResult):
                item = item.data
            if "num_query_matches" in item:
                results["evaluation"]["num_query_matches"].append(
                    item["num_query_matches"]
                )

    def evaluate_num_candidate_networks(self, results, **kwargs):
        results.setdefault("evaluation", {})
        results["evaluation"]["num_candidate_networks"] = []
        for item in results["results"]:
            if isinstance(item, LatheResult):
                item = item.data
            if "num_candidate_networks" in item:
                results["evaluation"]["num_candidate_networks"].append(
                    item["num_candidate_networks"]
                )

    def get_relevant_position(self, items, ground_truth):
        for i, item in enumerate(items):
            if item == ground_truth:
                return i + 1
        return -1

    def get_mean_reciprocal_rank(self, relevant_positions):
        if len(relevant_positions) == 0:
            return 0
        sum = 0
        for relevant_position in relevant_positions:
            if relevant_position != -1:
                reciprocal_rank = 1 / relevant_position
            else:
                reciprocal_rank = 0
            sum += reciprocal_rank

        mrr = sum / len(relevant_positions)
        return mrr

    def get_precision_at_k(self, k, relevant_positions):
        if len(relevant_positions) == 0:
            return 0
        sum = 0
        for relevant_position in relevant_positions:
            if relevant_position <= k and relevant_position != -1:
                sum += 1
        presition_at_k = sum / len(relevant_positions)
        return presition_at_k



import shelve
from copy import deepcopy
from os import makedirs
from os.path import dirname


class SchemaIndex:
    def __init__(self):
        self._dict = {}

    def __iter__(self):
        yield from self._dict.keys()

    def __getitem__(self, word):
        return self._dict[word]

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __repr__(self):
        return repr(self._dict)

    def __str__(self):
        return str(self._dict)

    def keys(self):
        yield from self.__iter__()

    def items(self):
        for key in self.__iter__():
            yield key, self.__getitem__(key)

    def values(self):
        for key in self:
            yield self[key]

    def create_entries(self, table_attributes, metrics):
        for table, attributes in table_attributes:
            self._dict.setdefault(table, {})
            for attribute in attributes:
                self._dict[table].setdefault(attribute, deepcopy(metrics))

    def tables_attributes(self):
        return {
            (table, attribute) for table in self for attribute in self[table]
        }

    def get_num_total_attributes(self):
        return sum([len(attribute) for attribute in self.values()])

    def persist_to_file(self, filename):
        makedirs(dirname(filename), exist_ok=True)
        with shelve.open(filename) as storage:
            for key, value in self._dict.items():
                storage[key] = value

    def load_from_file(self, filename):
        self._dict = {}

        with shelve.open(filename, flag="r") as storage:
            for key, value in storage.items():
                self[key] = value



from itertools import groupby
from typing import Generator, List, Tuple
from xml.sax import handler

import pandas as pd
import psycopg
from prettytable import PrettyTable
from psycopg.rows import dict_row
from psycopg.sql import SQL, Identifier
from sqlalchemy import create_engine

from polly.sources.postgres.iter import PostgresIter


class PostgresHandler:
    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        password: str,
        database: str,
        **kwargs,
    ):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.database_client = psycopg.connect(
            conninfo=f"postgresql://{user}:{password}@{host}:{port}/{database}",
            # row_factory=dict_row,
        )

    def get_tables_and_attributes(self):
        with self.database_client.cursor() as cur:
            sql = """
                    SELECT table_name,column_name
                    FROM information_schema.columns
                    WHERE table_schema='public'
                    ORDER by 1,2;
                    """
            cur.execute(sql)
            tables_attributes = {
                table: [attribute for table, attribute in group]
                for table, group in groupby(cur.fetchall(), lambda x: x[0])
            }

        return tables_attributes

    def get_tables_and_attributes_with_values(self):
        tables_attributes = self.get_tables_and_attributes()

        with self.database_client.cursor() as cur:
            for table, attributes in tables_attributes.items():
                for query in self.convert_dict_to_sql_queries(
                    table, attributes
                ):
                    print(f"Query: {query}")
                    cur.execute(query)
                    for row in cur.fetchall():
                        print(row)

        return tables_attributes

    def iterate_over_keywords(self, schema_index, **kwargs):
        database_table_columns = schema_index.tables_attributes()
        return PostgresIter(self.config, database_table_columns, **kwargs)

    def get_fk_constraints(self):
        with self.database_client.cursor() as cur:
            sql = """
                SELECT conname AS constraint_name,
                conrelid::regclass AS table_name,
                ta.attname AS column_name,
                confrelid::regclass AS foreign_table_name,
                fa.attname AS foreign_column_name
                FROM (
                    SELECT conname, conrelid, confrelid,
                        unnest(conkey) AS conkey, unnest(confkey) AS confkey
                    FROM pg_constraint
                    WHERE contype = 'f'
                ) sub
                JOIN pg_attribute AS ta ON ta.attrelid = conrelid AND ta.attnum = conkey
                JOIN pg_attribute AS fa ON fa.attrelid = confrelid AND fa.attnum = confkey
                ORDER BY 1,2,4;
            """
            cur.execute(sql)

            fk_constraints = {}
            for row in cur.fetchall():
                (
                    constraint,
                    table,
                    attribute,
                    foreign_table,
                    foreign_attribute,
                ) = (column.strip('"') for column in row)
                fk_constraints.setdefault(
                    (constraint, table), (table, foreign_table, [])
                )[2].append((attribute, foreign_attribute))

            for constraint in fk_constraints:
                table, foreign_table, attribute_mappings = fk_constraints[
                    constraint
                ]
                sql = """
                SELECT NOT EXISTS (
                    SELECT COUNT(t1.*), {}
                    FROM {} t1
                    GROUP BY {}
                    HAVING COUNT(t1.*)>1
                )
                """
                attributes_param = SQL(", ").join(
                    Identifier(attribute)
                    for attribute, foreign_attribute in attribute_mappings
                )
                sql = SQL(sql).format(
                    attributes_param,
                    Identifier(table),
                    attributes_param,
                )
                cur.execute(sql)
                cardinality = "1:1" if cur.fetchone()[0] else "N:1"
                fk_constraints[constraint] = (
                    cardinality,
                    table,
                    foreign_table,
                    attribute_mappings,
                )

        return fk_constraints

    def exec_sql(self, sql, **kwargs):
        show_results = kwargs.get("show_results", True)

        with self.database_client.cursor() as cur:
            try:
                cur.execute(sql)

                if cur.description:
                    table = PrettyTable(max_table_width=200)
                    table.field_names = [
                        f"{i}.{col[0]}" for i, col in enumerate(cur.description)
                    ]
                    for row in cur.fetchall():
                        table.add_row(row)
                if show_results:
                    print(table)
            except Exception as e:
                print(e)
                raise
            return table

    def execute(self, query: str):
        with self.database_client.cursor() as cur:
            cur.execute(query)
            return cur.fetchall()

    def get_dataframe(self, sql, **kwargs):
        engine = create_engine(
            "postgresql+psycopg2://{user}:{password}@{host}:5432/{database}".format(
                **self.config.connection
            )
        )
        df = pd.read_sql_query(sql, engine)
        return df

    def exist_results(self, sql):
        sql = f"SELECT EXISTS ({sql.rstrip(';')});"
        with self.database_client.cursor() as cur:
            try:
                cur.execute(sql)
                return cur.fetchone()[0]
            except Exception as e:
                print(e)
                return False

        return None

    def get_single_sample_per_table_query(
        self, table: str, attributes: List[str]
    ):
        query = f"SELECT {', '.join(attributes)} FROM {table} LIMIT 1;"

        return query

    def get_description_example_values(
        self,
    ) -> Generator[Tuple[str, List[Tuple[str, str]]], None, None]:
        tables_attributes = self.get_tables_and_attributes()

        for table, attributes in tables_attributes.items():
            filtered_attributes = [
                attr
                for attr in attributes
                if "__search_id" not in attr and "tsvector" not in attr
            ]
            sample_query = self.get_single_sample_per_table_query(
                table, filtered_attributes
            )
            sample_query_result = self.execute(sample_query)[0]

            yield table, list(zip(filtered_attributes, sample_query_result))


if __name__ == "__main__":
    handler = PostgresHandler(
        "localhost",
        5432,
        "polly",
        "pollypostgres",
    )

    print(
        handler.exec_sql(
            "SELECT table_schema, table_name, column_name FROM information_schema.columns ORDER by 1,2;"
        )
    )



import json
from collections import Counter
from pprint import pprint as pp
from queue import deque

from numpy import array, char
from sentence_transformers import util
from sklearn.metrics.pairwise import euclidean_distances

from polly.core.classes.keyword_match import KeywordMatch
from polly.sources.enum import DataSourceType
from polly.sources.mongo.utils.attribute_tree_generator import (
    AttributeTreeGenerator,
)
from polly.sources.mongo.utils.functions import load_datasets_ids_from_json
from polly.sources.mongo.utils.query_builder import MongoQueryBuilder
from polly.utils import (
    Graph,
    sort_dataframe_by_bow_size,
    sort_dataframe_by_token_length,
)


class CandidateNetwork(Graph):

    def __init__(
        self,
        graph_dict=None,
        data_source_type: DataSourceType = None,
    ):
        self.score = None
        self.__root = None
        self._df = None
        self.data_source_type = data_source_type

        super().__init__(graph_dict)

        if len(self) > 0:
            self.set_root()

    def get_root(self):
        return self.__root

    def set_root(self, vertex=None):
        if len(self) == 0:
            return None

        if vertex is not None:
            keyword_match, _ = vertex
            if not keyword_match.is_free():
                self.__root = vertex
                return vertex
            else:
                return None
        else:
            for candidate in self.vertices():
                keyword_match, _ = candidate
                if not keyword_match.is_free():
                    self.__root = candidate
                    return candidate

        print("The root of a Candidate Network cannot be a Keyword-Free Match.")
        print(self)
        raise ValueError(
            "The root of a Candidate Network cannot be a Keyword-Free Match."
        )

        # return None

    def reset_root(self, to_value=True):
        if to_value:
            for leaf in self.leaves():
                keyword_match, _ = leaf
                if keyword_match.has_value_mappings():
                    self.__root = leaf

    # def get_starting_vertex(self):
    #     if len(self) == 0:
    #         return None

    #     if self.__root is None:
    #         self.set_root()
    #     return self.__root

    # def get_starting_vertex(self):
    #     vertex = None
    #     for vertex in self.vertices():
    #         keyword_match,alias = vertex
    #         if not keyword_match.is_free():
    #             break
    #     return vertex

    def add_vertex(self, vertex):
        results = super().add_vertex(vertex)
        if self.__root is None:
            self.set_root(vertex)
        return results

    def add_keyword_match(self, heyword_match, **kwargs):
        alias = kwargs.get("alias", "t{}".format(self.__len__() + 1))
        vertex = (heyword_match, alias)
        return self.add_vertex(vertex)

    def add_adjacent_keyword_match(
        self, parent_vertex, keyword_match, edge_direction=">", **kwargs
    ):
        child_vertex = self.add_keyword_match(keyword_match, **kwargs)
        self.add_edge(
            parent_vertex, child_vertex, edge_direction=edge_direction
        )

    def keyword_matches(self):
        # return {keyword_match for keyword_match,alias in self.vertices()}
        for keyword_match, _ in self.vertices():
            yield keyword_match

    def non_free_keyword_matches(self):
        # return {keyword_match for keyword_match,alias in self.vertices() if not keyword_match.is_free()}
        for keyword_match, _ in self.vertices():
            if not keyword_match.is_free():
                yield keyword_match

    def num_free_keyword_matches(self):
        i = 0
        for keyword_match, _ in self.vertices():
            if keyword_match.is_free():
                i += 1
        return i

    def reciprocal_neighbors(self, schema_graph, vertex):
        """
        This function iterates over incoming neighbors which, in practice,
        also behaves like an outgoing neighbor. That is, the foreign key
        is unique (not explicitly), which sugests a cardinality of 1:1
        between the two tables.
        """
        keyword_match, _ = vertex
        for neighbor_vertex in self.incoming_neighbors(vertex):
            neighbor_km, _ = neighbor_vertex

            edge_info = schema_graph.get_edge_info(
                neighbor_km.table, keyword_match.table
            )
            for _, (cardinality, _) in edge_info.items():
                if cardinality == "1:1":
                    yield neighbor_vertex

    def is_sound(self, schema_graph):
        if len(self) < 3:
            return True

        def has_duplicate_tables(keyword_match, neighbors, direction=">"):
            # print(f'keyword_match:\n{keyword_match}\nneighbors:\n{neighbors}\ndirection:{direction}\n')

            if len(neighbors) >= 2:
                neighbor_tables = Counter()
                fkeys_tables = Counter()

                for neighbor_km, _ in neighbors:
                    neighbor_tables[neighbor_km.table] += 1

                    if neighbor_km.table not in fkeys_tables:
                        if direction == ">":
                            edge_info = schema_graph.get_edge_info(
                                keyword_match.table, neighbor_km.table
                            )
                        elif direction == "<":
                            edge_info = schema_graph.get_edge_info(
                                neighbor_km.table, keyword_match.table
                            )

                        fkeys_tables[neighbor_km.table] = len(edge_info)

                return len(neighbor_tables - fkeys_tables) > 0

            return False

        for vertex in self.vertices():
            km, _ = vertex

            # check if there is a case A->B<-C, when A.table==C.table
            outgoing_neighbors = list(self.outgoing_neighbors(vertex))
            if has_duplicate_tables(km, outgoing_neighbors):
                return False

            # check if there is a case A<-B->C, when A.table==C.table and
            # there is no duplicated foreign key in B in the database instance
            reciprocal_neighbors = list(
                self.reciprocal_neighbors(schema_graph, vertex)
            )
            if has_duplicate_tables(km, reciprocal_neighbors, direction="<"):
                return False
        return True

    def remove_vertex(self, vertex):
        print("vertex:\n{}\n_graph_dict\n{}".format(vertex, self._graph_dict))
        _, incoming_neighbors = self._graph_dict[vertex]
        for neighbor in incoming_neighbors:
            self._graph_dict[neighbor][0].remove(vertex)
        self._graph_dict.pop(vertex)

    def is_total(self, query_match):
        return set(self.non_free_keyword_matches()) == set(query_match)

    def contains_keyword_free_match_leaf(self):
        for vertex in self.leaves():
            keyword_match, _ = vertex
            if keyword_match.is_free():
                return True
        return False

    def minimal_cover(self, query_match):
        return (
            self.is_total(query_match)
            and not self.contains_keyword_free_match_leaf()
        )

    def unaliased_edges(self):
        for (keyword_match, _), (neighbor_keyword_match, _) in self.edges():
            yield (keyword_match, neighbor_keyword_match)

    def get_sentences(
        self, schema_graph, schema_index, database_handler, **kwargs
    ):
        cjn_ranking_method = kwargs.get("cjn_ranking_method", 0)
        snapshot_method = kwargs.get("snapshot_method", "first")
        snapshot_size = kwargs.get("snapshot_size", 512)

        if self._df is None:
            sql = self.get_sql_from_cn(
                schema_graph,
                schema_index,
                show_table_alias=True,
                show_only_indexable_attributes=True,
            )
            self._df = database_handler.get_dataframe(sql)

        df = self._df.copy()

        if snapshot_method == "max_length":
            df = sort_dataframe_by_token_length(df)
        elif snapshot_method == "bow_size":
            df = sort_dataframe_by_bow_size(df)
        elif snapshot_method == "random":
            df = df.sample(n=min(snapshot_size, len(df)))

        df = df.head(snapshot_size)

        sentences = []

        if cjn_ranking_method <= 2:
            for i, (index, row) in enumerate(df.iterrows()):
                attrs_repr = []
                for km, alias in self.vertices():
                    for km_type, table, attr, keywords in km.mappings():
                        if attr != "*":
                            value = row[f"{alias}.{attr}"]
                            attrs_repr.append(f"{table}.{attr}: {value}")
                        else:
                            attrs_repr.append(f'{table}: {" ".join(keywords)}')

                sentence = " | ".join(attrs_repr)
                sentences.append(sentence)

            if cjn_ranking_method == 2:
                sentences = [" || ".join(sentences)]

        if cjn_ranking_method == 3:
            attrs_repr = []
            for km, alias in self.vertices():
                for km_type, table, attr, keywords in km.mappings():
                    if attr == "*":
                        attrs_repr.append(f'{table}: {" ".join(keywords)}')
                    else:
                        # numpy.char
                        value = ", ".join(
                            char.mod("%s", df[f"{alias}.{attr}"].values)
                        )
                        attrs_repr.append(f"{table}.{attr}: {value}")
            sentence = " | ".join(attrs_repr)
            sentences.append(sentence)

        return sentences

    def calculate_score(
        self,
        query_match,
        keyword_query,
        schema_graph,
        schema_index,
        database_handler,
        bert_model,
        **kwargs,
    ):
        cjn_ranking_method = kwargs.get("cjn_ranking_method", 0)
        bert_similarity = kwargs.get("bert_similarity", "cos")
        snapshot_method = kwargs.get("snapshot_method", "first")
        snapshot_size = kwargs.get("snapshot_size", 512)

        if cjn_ranking_method == 0:
            self.score = query_match.total_score / len(self)
            return self.score

        sentences_to_compare = [keyword_query] + self.get_sentences(
            schema_graph, schema_index, database_handler, **kwargs
        )

        embeddings = bert_model.encode(array(sentences_to_compare))

        query_emb = array([embeddings[0]])
        row_emb = embeddings[1:]
        if bert_similarity == "cos":
            scores = util.cos_sim(query_emb, row_emb)
        elif bert_similarity == "dot":
            scores = util.dot_score(query_emb, row_emb)
        else:
            scores = euclidean_distances(query_emb, row_emb)
        self.score = scores.mean().item()
        return self.score

    def get_qm_score(self):
        return self.score * len(self)

    def __eq__(self, other):

        if not isinstance(other, CandidateNetwork):
            return False

        self_root_km, _ = self.__root
        # if other.get_root() is None:
        #     print(f'OTHER ROOT IS NONE')
        #     print(other)
        other_root_km, _ = other.get_root()

        other_hash = None

        if self_root_km == other_root_km:
            other_hash = hash(other)
        else:
            for keyword_match, alias in other.vertices():
                if self_root_km == keyword_match:
                    root = (keyword_match, alias)
                    # print(f'Root: {root}')
                    other_hash = other.hash_from_root(root)

        if other_hash is None:
            return False

        # print(f'Equal:{hash(self)==other_hash}\nSelf:\n{self}\nOther:\n{other}\n\n\n')

        return hash(self) == other_hash

    def __hash__(self):
        if len(self) == 0:
            return hash(None)
        if self.__root is None:
            self.set_root()
        return self.hash_from_root(self.__root)

    def hash_from_root(self, root):
        hashable = []

        level = 0
        visited = set()

        queue = deque()
        queue.append((level, root))

        while queue:
            level, vertex = queue.popleft()
            keyword_match, alias = vertex
            children = Counter()
            visited.add(alias)

            for adj_vertex in self.neighbors(vertex):
                adj_km, adj_alias = adj_vertex
                if adj_alias not in visited:
                    queue.append((level + 1, adj_vertex))
                    children[adj_km] += 1

            if len(hashable) < level + 1:
                hashable.append(set())

            hashable[level].add((keyword_match, frozenset(children.items())))

        hashcode = hash(tuple(frozenset(items) for items in hashable))
        return hashcode

    def __repr__(self):
        if len(self) == 0:
            return "EmptyCN"
        print_string = [
            "\t" * level + direction + str(vertex[0])
            for direction, level, vertex in self.leveled_dfs_iter()
        ]
        return "\n".join(print_string)

    def to_json_serializable(self):
        return [
            {
                "keyword_match": keyword_match.to_json_serializable(),
                "alias": alias,
                "outgoing_neighbors": [
                    alias for (km, alias) in outgoing_neighbors
                ],
                "incoming_neighbors": [
                    alias for (km, alias) in incoming_neighbors
                ],
            }
            for (keyword_match, alias), (
                outgoing_neighbors,
                incoming_neighbors,
            ) in self._graph_dict.items()
        ]

    def to_json(self):
        return json.dumps(self.to_json_serializable())

    @staticmethod
    def from_json_serializable(json_serializable_cn):
        alias_hash = {}
        edges = []
        for vertex in json_serializable_cn:
            keyword_match = KeywordMatch.from_json_serializable(
                vertex["keyword_match"],
                vertex["keyword_match"]["database"],
            )
            alias_hash[vertex["alias"]] = keyword_match

            for outgoing_neighbor in vertex["outgoing_neighbors"]:
                edges.append((vertex["alias"], outgoing_neighbor))

        candidate_network = CandidateNetwork()
        for alias, keyword_match in alias_hash.items():
            candidate_network.add_vertex((keyword_match, alias))
        for alias1, alias2 in edges:
            vertex1 = (alias_hash[alias1], alias1)
            vertex2 = (alias_hash[alias2], alias2)
            candidate_network.add_edge(vertex1, vertex2)

        return candidate_network

    @staticmethod
    def from_json(json_cn):
        return CandidateNetwork.from_json_serializable(json.loads(json_cn))

    def get_sql_from_cn(self, schema_graph, schema_index, **kwargs):
        rows_limit = kwargs.get("rows_limit", 1000)
        show_evaluation_fields = kwargs.get("show_evaluation_fields", False)
        tsvector_field_suffix = kwargs.get("tsvector_field_suffix", "_tsvector")
        show_table_alias = kwargs.get("show_table_alias", False)
        distinct = kwargs.get("distinct", True)
        show_only_indexable_attributes = kwargs.get(
            "show_only_indexable_attributes", False
        )

        hashtables = {}  # used for disambiguation
        used_fks = {}

        selected_attributes = set()
        filter_conditions = []
        disambiguation_conditions = []
        selected_tables = []

        tables__search_id = []
        relationships__search_id = []

        def get_column_type(table, attribute):
            # TODO: Handle attributes from the "varchar"  domain
            # but do not have to be indexed
            if table == "title" and attribute == "production_year":
                return "integer"
            if table == "movie" and attribute == "year":
                return "integer"
            if table == "organization" and attribute == "abbreviation":
                return "varchar"
            if attribute == "stars":
                return "integer"
            return "fulltext_indexed"

        for prev_vertex, direction, vertex in self.dfs_pair_iter(
            root_predecessor=True
        ):
            keyword_match, alias = vertex
            for type_km, table, attribute, keywords in keyword_match.mappings():
                attributes_to_select = [(alias, attribute)]
                if show_only_indexable_attributes and attribute == "*":
                    attributes_to_select = [
                        (alias, attr) for attr in schema_index[table].keys()
                    ]
                for (
                    alias_to_select,
                    attribute_to_select,
                ) in attributes_to_select:
                    selected_attr = f"{alias_to_select}.{attribute_to_select}"
                    if show_table_alias:
                        selected_attr = f'{selected_attr} AS "{alias_to_select}.{attribute_to_select}"'
                    selected_attributes.add(selected_attr)

                sql_keywords = [
                    keyword.replace("'", "''") for keyword in keywords
                ]

                if type_km == "v":

                    column_type = get_column_type(
                        keyword_match.table, attribute
                    )

                    if column_type == "fulltext_indexed":
                        condition = f"{alias}.{attribute}{tsvector_field_suffix} @@ to_tsquery('{ ' & '.join(sql_keywords) }')"
                        filter_conditions.append(condition)
                    else:
                        for sql_keyword in sql_keywords:
                            if column_type == "varchar":
                                condition = f"{alias}.{attribute} ILIKE '%{sql_keyword}%'"
                            elif column_type == "integer":
                                condition = (
                                    f"{alias}.{attribute} = {sql_keyword}"
                                )
                            else:
                                condition = f"CAST({alias}.{attribute} AS VARCHAR) ILIKE '%{sql_keyword}%'"
                            filter_conditions.append(condition)

            hashtables.setdefault(keyword_match.table, []).append(alias)

            if show_evaluation_fields:
                tables__search_id.append(f"{alias}.__search_id")

            if prev_vertex is None:
                selected_tables.append(f'"{keyword_match.table}" {alias}')
            else:
                # After the second table, it starts to use the JOIN syntax
                _, prev_alias = prev_vertex
                if direction == ">":
                    constraint_keyword_match, constraint_alias = prev_vertex
                    foreign_keyword_match, foreign_alias = vertex
                else:
                    constraint_keyword_match, constraint_alias = vertex
                    foreign_keyword_match, foreign_alias = prev_vertex

                edge_info = schema_graph.get_edge_info(
                    constraint_keyword_match.table, foreign_keyword_match.table
                )

                for constraint in edge_info:
                    if constraint not in used_fks.setdefault(
                        constraint_alias, []
                    ):
                        used_fks[constraint_alias].append(constraint)

                        _, attribute_mappings = edge_info[constraint]

                        join_conditions = []
                        for (
                            constraint_column,
                            foreign_column,
                        ) in attribute_mappings:
                            join_conditions.append(
                                f"{constraint_alias}.{constraint_column} = {foreign_alias}.{foreign_column}"
                            )
                        txt_join_conditions = " AND ".join(join_conditions)
                        selected_tables.append(
                            f' JOIN "{keyword_match.table}" {alias} ON {txt_join_conditions}'
                        )
                        break

                if show_evaluation_fields:
                    relationships__search_id.append(
                        f"({alias}.__search_id, {prev_alias}.__search_id)"
                    )

        for _, aliases in hashtables.items():
            for i in range(len(aliases)):
                for j in range(i + 1, len(aliases)):
                    disambiguation_conditions.append(
                        f"{aliases[i]}.ctid <> {aliases[j]}.ctid"
                    )

        if len(tables__search_id) > 0:
            tables__search_id = [f"({', '.join(tables__search_id)}) AS Tuples"]
        if len(relationships__search_id) > 0:
            txt_relationships = ", ".join(relationships__search_id)
            relationships__search_id = [
                f"({txt_relationships}) AS Relationships"
            ]

        where_conditions = disambiguation_conditions + filter_conditions
        if len(where_conditions) > 0:
            where_clause = "WHERE {}".format(" AND ".join(where_conditions))
        else:
            where_clause = ""

        sql_text = "\nSELECT {} {} FROM {} {} LIMIT {};".format(
            " DISTINCT " if distinct else "",
            ", ".join(
                tables__search_id
                + relationships__search_id
                + list(selected_attributes)
            ),
            " ".join(selected_tables),
            where_clause,
            rows_limit,
        )
        return sql_text

    def generate_pipeline(self):
        self.reset_root(to_value=True)

        if len(self) == 1:
            yield None, self.__root
        else:
            for prev_vertex, direction, vertex in self.dfs_pair_iter(
                root_predecessor=True, rev=False, start_vertex=self.__root
            ):
                if prev_vertex == None:
                    continue

                keyword_match, alias = vertex
                yield prev_vertex, vertex

    def get_mongo_query_from_cn(self, database_name, schema_graph):
        selected_attributes = set()
        selected_tables = []

        next_step_lookup_field = None
        last_added_collection = None

        datasets_collection_ids_dict = load_datasets_ids_from_json()

        query = MongoQueryBuilder(database_name)
        self.reset_root(to_value=True)

        fields_to_project = []

        for prev_vertex, direction, vertex in self.dfs_pair_iter(
            root_predecessor=True, rev=False, start_vertex=self.__root
        ):
            keyword_match, alias = vertex
            project_fields = []
            value_fields = []

            tree_generator = AttributeTreeGenerator()

            for type_km, _, attribute, keywords in keyword_match.mappings():
                selected_attributes.add(f"{alias}.{attribute}")
                if type_km == "v":
                    for keyword in keywords:
                        if isinstance(keyword, str):
                            sql_keyword = keyword.replace("'", "''")
                            value_fields.append((attribute, sql_keyword))
                        elif isinstance(keyword, frozenset):
                            value_fields.append((attribute, keyword))
                    continue

            if prev_vertex is None:
                query.set_starting_collection(keyword_match.table)
                selected_tables.append(f"{keyword_match.table} {alias}")
                edge_info = schema_graph.get_edge_info(
                    keyword_match.table, keyword_match.table
                )
                if edge_info:
                    for constraint in edge_info:
                        _, attribute_mappings = edge_info[constraint]
                        if isinstance(attribute_mappings, list):
                            l = []
                            l.append(attribute_mappings)
                            attribute_mappings = l
                        for (
                            constraint_column,
                            foreign_column,
                        ) in attribute_mappings:
                            project_fields.append(constraint_column)
                for value_field in value_fields:
                    tree_generator.add_attribute(value_field[0], value_field[1])
                query.new_match_v3(
                    keyword_match.table, tree_generator.generate()
                )
            else:
                _, prev_alias = prev_vertex
                if direction == ">":
                    constraint_keyword_match, constraint_alias = prev_vertex
                    foreign_keyword_match, foreign_alias = vertex
                else:
                    constraint_keyword_match, constraint_alias = vertex
                    foreign_keyword_match, foreign_alias = prev_vertex

                edge_info = schema_graph.get_edge_info(
                    constraint_keyword_match.table, foreign_keyword_match.table
                )

                for constraint in edge_info:
                    _, attribute_mappings = edge_info[constraint]

                    if type(attribute_mappings) is list:
                        l = []
                        l.append(attribute_mappings)
                        attribute_mappings = l

                    if last_added_collection == keyword_match.table:
                        continue

                    if not next_step_lookup_field:
                        constraint_column, foreign_column = attribute_mappings[
                            0
                        ]
                        query.lookup(
                            constraint_column,
                            foreign_column,
                            keyword_match.table,
                            keyword_match.table,
                        )
                    else:
                        constraint_column, foreign_column = attribute_mappings[
                            0
                        ]
                        query.lookup(
                            f"{next_step_lookup_field}.{constraint_column}",
                            foreign_column,
                            keyword_match.table,
                            keyword_match.table,
                        )

                if last_added_collection != keyword_match.table:
                    query.unwind(keyword_match.table, insert=True)
                    if value_fields:
                        for attribute, sql_keyword in value_fields:
                            # attribute_name = (
                            #     keyword_match.table + "." + attribute
                            # )
                            tree_generator.add_attribute(
                                attribute,
                                sql_keyword,
                            )
                        query.new_match_v3(
                            keyword_match.table, tree_generator.generate()
                        )
                        next_step_lookup_field = keyword_match.table
                    else:
                        next_step_lookup_field = keyword_match.table
                else:
                    for attribute, sql_keyword in value_fields:
                        tree_generator.add_attribute(attribute, sql_keyword)
                    query.new_match_v3(
                        keyword_match.table, tree_generator.generate()
                    )

            if next_step_lookup_field:
                if (
                    keyword_match.table
                    in datasets_collection_ids_dict[database_name]
                ):
                    for attribute in datasets_collection_ids_dict[
                        database_name
                    ][keyword_match.table]:
                        fields_to_project.append(
                            next_step_lookup_field + "." + attribute,
                        )
            else:
                fields_to_project.append(
                    datasets_collection_ids_dict[database_name][
                        keyword_match.table
                    ][0]
                )

            last_added_collection = keyword_match.table

        # query.project(
        #     fields_to_project,
        # )

        return query



from os import walk

import polars as plrs
import vaex
from pylathedb.handlers.base import Iter
from pylathedb.utils.tokenizer import Tokenizer


class CSVIter(Iter):

    def __init__(
        self, config, attributes_to_fetch, tokenizer: Tokenizer, **kwargs
    ):
        self.config = config
        self.attributes_to_fetch = attributes_to_fetch
        self.data_path_hash = {}
        self.tokenizer = tokenizer
        self.chunk_size = kwargs.get("read_chunk_size", 20_000)

        self._get_indexable_schema_elements()

    def _get_indexable_schema_elements(
        self,
    ):  # , tables_attributes: dict[str, list[str]]
        for dirpath, _, filenames in walk(self.config.data_source_path):
            print(f"Current path: {dirpath}")

            if len(filenames):
                for filename in filenames:
                    file = dirpath + filename
                    # print(f'Indexing file: {file}')

                    # Read header
                    header = plrs.scan_csv(file).columns

                    self.data_path_hash[file] = header

        # for data_path, attributes in tables_attributes.items():
        #   if data_path not in self.attributes_to_fetch:
        #     print('Data Path not listed in configuration, ignoring...')
        #     continue

        #   for attribute in attributes:
        #     if attribute in self.attributes_to_fetch[data_path]:
        #       if data_path not in data_path_hash:
        #         data_path_hash[data_path] = [attribute]
        #       else:
        #         data_path_hash[data_path].append(attribute)

        # self.data_path_hash = data_path_hash

    def iterate(self):
        # print(f'Data Path Hash: {self.data_path_hash}')
        total_items = len(list(self.data_path_hash.keys()))
        processed_items = 0
        print(f"Total items for indexing: {total_items}")
        for data_path, attributes in self.data_path_hash.items():
            print(f"Currently in Data Path: {data_path}")
            try:
                csv_reader = vaex.from_csv(
                    data_path, chunk_size=self.chunk_size
                )
            except Exception as e:
                csv_reader = vaex.from_csv(
                    data_path, chunk_size=self.chunk_size, delimiter=";"
                )
            global_data_path_row_index = 0

            attributes = [
                attribute for attribute in attributes if attribute != ""
            ]

            # print(f'Current attributes for processing: {attributes})')

            for df in csv_reader:
                # print('Reading chunk from data path...')
                projected_df = df[attributes]

                for _, row in projected_df.iterrows():
                    global_data_path_row_index += 1
                    for row_item in row:
                        content = str(row[row_item])
                        tokens = self.tokenizer.tokenize(content)

                        for token in tokens:
                            # print(data_path, global_data_path_row_index, row_item, token)
                            yield data_path, global_data_path_row_index, row_item, token

            processed_items += 1
            print(f"Processed {processed_items} of {total_items}")


if __name__ == "__main__":
    input_data = {
        "./files/25kjul12.csv": [
            "Department family",
            "Entity",
            "Expense type",
            "Expense area",
            "Supplier",
        ],
        "./files/25kjun10.csv": [
            "Department family",
            "Entity",
            "Expense type",
            "Expense area",
            "Supplier",
        ],
    }

    i = CSVIter(input_data, Tokenizer())
    i.data_path_hash = input_data
    for r in i.iterate():
        print(r)



import json
from typing import Dict, List

from openai import OpenAI


class LLMChat:
    def __init__(
        self,
        api_key: str,
        model: str = "gpt-4o",
        temperature: float = 0.0,
        top_p: float = 1.0,
        show_response: bool = False,
    ):
        self.api = OpenAI(api_key=api_key)
        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.show_response = show_response

    def chat(
        self,
        messages: List[Dict[str, str]],
        cast_response_to_json: bool = False,
    ):
        response_data = self.api.chat.completions.create(
            model=self.model,
            messages=messages,
            temperature=self.temperature,
            top_p=self.top_p,
        )

        response = response_data.choices[0].message.content

        if self.show_response:
            print(response)

        if cast_response_to_json:
            response = json.loads(response)

        return response

    def build_system_message(self, system_prompt: str):
        return {"role": "system", "content": system_prompt.replace("\n", "")}

    def build_user_message(self, user_prompt: str):
        return {"role": "user", "content": user_prompt.replace("\n", "")}



import shelve
from os import makedirs
from os.path import dirname
from random import sample

from polly.index.babel.hash import BabelHash
from polly.utils import calculate_iaf, calculate_inverse_frequency, calculate_tf


class ValueIndex:

    def __init__(self, **kwargs):
        self._dict = {}
        self.persistant_filename = kwargs.get("persistant_filename", None)
        self.update_cache = kwargs.get("update_cache", True)

    def __contains__(self, keyword):
        if keyword in self._dict:
            return True

        if self.persistant_filename:
            with shelve.open(self.persistant_filename, flag="r") as storage:
                if keyword in storage:
                    return True

        return False

    def _set_underlying_item(self, keyword, value):
        self._dict[keyword] = value

    def _get_underlying_item(self, keyword):
        item = None
        if keyword in self._dict:
            item = self._dict[keyword]
        elif self.persistant_filename:
            with shelve.open(self.persistant_filename, flag="r") as storage:
                item = storage.get(keyword, None)
                if self.update_cache:
                    self._set_underlying_item(keyword, item)
        if item is None:
            item = (0, BabelHash())
            # raise KeyError('The keyword {keyword} is not in the ValueIndex.')
        return item

    def __getitem__(self, keyword):
        _inverse_frequency, value = self._get_underlying_item(keyword)
        return value

    def __setitem__(self, keyword, value):
        raise Exception(
            "Invalid operation. ValueIndex items shoud be set using the add_mapping method."
        )

    def __iter__(self):
        # This method only iterates over the keywords in memory
        yield from self._dict.keys()

    def keys(self):
        yield from self.__iter__()

    def items(self):
        for key in self.__iter__():
            yield key, self.__getitem__(key)

    def get_mappings(self, keyword, table, attribute):
        return self[keyword][table][attribute]

    def add_mapping(self, keyword, table, attribute, ctid):
        # This method does not change the persistant_filename
        self._dict.setdefault(keyword, (0, BabelHash()))
        self[keyword].setdefault(table, BabelHash())
        self[keyword][table].setdefault(attribute, []).append(ctid)

    def get_inverse_frequency(self, keyword):
        inverse_frequency, _value = self._get_underlying_item(keyword)
        return inverse_frequency

    def set_inverse_frequency(self, keyword, inverse_frequency):
        # This method does not change the persistant_filename
        _old_inverse_frequency, old_value = self._dict[keyword]
        self._dict[keyword] = (inverse_frequency, old_value)

    def get_frequency(self, keyword, table, attribute):
        if table not in self or attribute not in self[table]:
            return 0

        return len(self[keyword][table][attribute])

    def frequencies(self):
        for word in self:
            inverse_frequency = self.get_inverse_frequency(word)
            for table in self[word]:
                for attribute, ctids in self[word][table].items():
                    frequency = len(ctids)
                    yield table, attribute, frequency, inverse_frequency

    def get_iaf(self, weight_scheme, keyword):
        return calculate_iaf(
            weight_scheme, inverse_frequency=self.get_inverse_frequency(keyword)
        )

    def get_tf(
        self, weight_scheme, keyword, table, attribute, max_frequency=None
    ):
        return calculate_tf(
            weight_scheme,
            self.get_frequency(keyword, table, attribute),
            max_frequency=max_frequency,
        )

    def __repr__(self):
        return f"<ValueIndex {repr(self._dict)}>"

    def persist_to_file(self, filename):
        makedirs(dirname(filename), exist_ok=True)
        with shelve.open(filename) as storage:
            storage["__babel__"] = BabelHash.babel
            for key, underlying_value in self._dict.items():
                storage[key] = underlying_value

    def load_from_file(self, persistant_filename, **kwargs):
        """
        The load_method specifies the set of keywords to me loaded, which can be:
        - keywords: A list or set of keywords to be loaded. If the list is empty,
        no keyword is loaded in this method.
        - sample: A sample of keywords from shelve is loaded. The number of keywords
        in the sample is defined by sample_size, which is 15 by default.
        - all_keywords: All the keywords from shelve are loaded. Beware that this
         load methods might be expensive.
        """
        self._dict = {}

        self.persistant_filename = persistant_filename
        load_method = kwargs.get("load_method", "keywords")
        keywords = kwargs.get("keywords", [])
        sample_size = kwargs.get("sample_size", 15)

        with shelve.open(self.persistant_filename, flag="r") as storage:

            if load_method != "keywords":
                keywords = storage.keys()

                if load_method == "sample":
                    keywords = sample(keywords, k=sample_size)

            if "__babel__" in storage:
                BabelHash.babel.update(storage["__babel__"])

            for keyword in keywords:
                if keyword == "__babel__":
                    continue
                try:
                    self._set_underlying_item(keyword, storage[keyword])
                except KeyError:
                    self._set_underlying_item(keyword, None)
                    # print(f'Keyword {keyword} not present in the ValueIndex.')
                    continue



import json
from pathlib import Path
from typing import List, Tuple

from polly.llm.chat import LLMChat
from polly.sources.enum import DataSourceType
from polly.sources.handler import SourceHandler


class LLMSchemaDescriber:
    __slots__ = ["llm_chat"]

    def __init__(self, llm_chat: LLMChat):
        self.llm_chat = llm_chat

    def generate_data_source_elements_example_snippet(
        self, values: List[Tuple[str, str]]
    ):
        value_snippets = []
        for key, value in values:
            value_snippets.append(f"{key}: {str(value)}")

        return "\n".join(value_snippets)

    def generate_data_source_elements_example_dict(
        self, values: List[Tuple[str, str]]
    ):
        value_dict = {}
        for key, value in values:
            value_dict[key] = str(value)

        return value_dict

    def generate_data_source_elements_description(
        self, table: str, values: List[Tuple[str, str]]
    ):
        columns = [key for key, _ in values]
        data_source_elements_snippet = (
            self.generate_data_source_elements_example_snippet(values)
        )

        print(f"Generating data source elements description for table: {table}")

        system_prompt = self._generate_table_excerpt_system_prompt(
            table,
            columns,
            data_source_elements_snippet,
        )
        system_message = self.llm_chat.build_system_message(system_prompt)

        user_prompt = self._generate_column_description_user_prompt()
        user_message = self.llm_chat.build_user_message(user_prompt)

        messages = [system_message, user_message]

        description = self.llm_chat.chat(messages, cast_response_to_json=True)

        print(f"Generating attribute description: {description}")

        return description

    def generate_table_description(
        self, table: str, values: List[Tuple[str, str]]
    ):
        columns = [key for key, _ in values]
        data_source_elements_snippet = (
            self.generate_data_source_elements_example_snippet(values)
        )

        print(f"Generating data source elements description for table: {table}")

        system_prompt = self._generate_table_excerpt_system_prompt(
            table,
            columns,
            data_source_elements_snippet,
        )

        system_message = self.llm_chat.build_system_message(system_prompt)

        user_prompt = self._generate_table_description_user_prompt()
        user_message = self.llm_chat.build_user_message(user_prompt)

        messages = [system_message, user_message]

        description = self.llm_chat.chat(messages, cast_response_to_json=True)

        print(f"Generating table description: {description}")

        return description

    def generate_table_and_column_description(
        self, table: str, values: List[Tuple[str, str]]
    ):
        columns = [key for key, _ in values]
        data_source_elements_description = (
            self.generate_data_source_elements_description(table, values)
        )

        table_description = self.generate_table_description(table, values)

        return {
            "table": table,
            "table_description": table_description["description"],
            "columns": columns,
            "column_description": data_source_elements_description[
                "column_description"
            ],
            "example": self.generate_data_source_elements_example_dict(values),
        }

    def _generate_table_excerpt_system_prompt(
        self, table: str, columns: List[str], data: str
    ) -> str:
        system_message = """Given the table {table}
                                  and its respective columns: {columns}
                                  Given the column values below:
                                  {data}""".format(
            table=table,
            columns=columns,
            data=data,
        )

        return system_message

    def _generate_column_description_user_prompt(self):
        user_prompt = """
            Task: Describe the information within a column from
            table using columns values, without itemization.

            Instructions:
            1. Look at the input given to you.
            2. Look at the column values in detail.
            3. Avoid citing the column values in the description.
            4. Describe the target columns.
            5. Return the description in plain text using the JSON
            schema below. Do not output markdown.

            "column_description": {
                "namecolumn1": "description1",
                "namecolumn2": "description2",
                "namecolumn3": "description3",
                "namecolumn4": "description4",
                "namecolumnN": "descriptionN"
            }
        """

        return user_prompt

    def _generate_table_description_user_prompt(self):
        user_prompt = """
            Task: Describe the information within a table in a concise manner.
            Do not comment about the columns or example values provided.

            Instructions:
            1. Look at the input given to you.
            2. Look at the column values in detail.
            3. Describe the table.
            4. Return the description in plain text using the JSON
            schema below. Do not output markdown.

            {
                "description": "description of the table"
            }
        """

        return user_prompt

    def generate_schema_description(self, source_handler: SourceHandler):
        schema_description = []
        from pprint import pprint as pp

        if source_handler.type != DataSourceType.DOCUMENT:
            return

        file_path = "/".join(["tmp", source_handler.database + ".json"])
        if not Path(file_path).exists():
            with open(file_path, mode="w") as f:
                print(
                    f"Creating file {file_path} with content: {schema_description}"
                )
                json.dump(schema_description, f)

        with open(
            file_path,
            "r",
        ) as f:
            schema_description = json.load(f)

            print("EXISTING SCHEMA DATA")
            print(schema_description)

        for (
            table,
            values,
        ) in source_handler.get_description_example_values():
            table_exists = [
                item for item in schema_description if item["table"] == table
            ]

            if len(table_exists) != 0:
                continue

            table_description = self.generate_table_and_column_description(
                table, values
            )

            print(f"Table {table} description generated")
            pp(table_description)

            schema_description.append(table_description)

            with open(file_path, "w") as f:
                json.dump(schema_description, f, indent=2)

        return schema_description



import json
from pathlib import Path
from pprint import pp
from typing import Dict, List

from polly.config import ConfigHandler
from polly.core.classes.candidate_network import CandidateNetwork
from polly.core.classes.keyword_match import KeywordMatch
from polly.core.handlers.candidate_network_handler import (
    CandidateNetworkHandler,
)
from polly.core.handlers.query_match_handler import QueryMatchHandler
from polly.index.handler import IndexHandler
from polly.llm.chat import LLMChat
from polly.llm.describer import LLMSchemaDescriber
from polly.llm.matcher import LLMKMatcher
from polly.llm.ranker import LLMQMRanker
from polly.sources.enum import DataSourceType
from polly.sources.handler import SourceHandler
from polly.utils import get_logger
from polly.utils.result import PollyResult


class Polly:
    source_handlers: Dict[str, SourceHandler] = {}
    index_handlers: Dict[str, IndexHandler] = {}

    def __init__(
        self,
        api_key,
        config_directory="./config/",
        **kwargs,
    ):
        # Application configurations
        self.config = ConfigHandler(
            config_directory,
        )
        self.config_directory = config_directory
        self.api_key = api_key

        # Runtime configurations
        self.top_ranked_qm_count = kwargs.get("top_ranked_qm_count", 100)
        self.max_qm_size = kwargs.get("max_qm_size", 3)
        self.top_ranked_result_count = kwargs.get(
            "top_ranked_result_count",
            10,
        )
        self.top_ranked_result_per_qm = kwargs.get(
            "top_ranked_result_per_qm", 2
        )
        self.max_result_size = kwargs.get("max_result_size", 4)
        self.topk_cjns_per_qm_list = kwargs.get("top_cjns_per_qm_list", [1])
        self.logger = get_logger("polly", self.config)

        # Setup
        self._init_source_handlers()
        self._init_index_handlers()

        self.llm_chat = LLMChat(
            self.api_key,
        )

    def execute_keyword_query(self, keyword_query: str):
        with open("./datasets_data.json") as f:
            tables_attributes = json.load(f)

        result = []

        for database in self.source_handlers:
            self.logger.info(
                f"Executing in database: {database} ({self.source_handlers[database].type.value})"
            )
            llm_matcher = LLMKMatcher(
                keyword_query=keyword_query,
                database=database,
                llm_chat=self.llm_chat,
            )

            match_result = llm_matcher.generate_keyword_matches()
            keyword_matches = match_result.keyword_matches

            qm_handler = QueryMatchHandler()
            query_matches = qm_handler.generate_query_matches(
                keyword_query.split(),
                keyword_matches,
            )

            ranker = LLMQMRanker(
                keyword_query=keyword_query,
                database=database,
                query_matches=query_matches,
                llm_chat=self.llm_chat,
            )

            ranked_qms = ranker.rank_query_matches()

            source_handler = self.source_handlers[database]
            cjn_handler = CandidateNetworkHandler(
                source_handler,
            )

            index_handler = IndexHandler(self.config, source_handler)
            index_handler.load_schema_graph()

            cjns: List[CandidateNetwork] = cjn_handler.generate_cns(
                tables_attributes,
                index_handler.schema_graph,
                ranked_qms,
                keyword_query.split(),
                keyword_query,
            )

            generated_query = None

            if len(cjns) > 0:
                if source_handler.type == DataSourceType.RELATIONAL:
                    generated_query = cjns[0].get_sql_from_cn(
                        index_handler.schema_graph, index_handler.schema_index
                    )
                elif source_handler.type == DataSourceType.DOCUMENT:
                    generated_query = cjns[0].get_mongo_query_from_cn(
                        database, index_handler.schema_graph
                    )

            database_result = {
                "keyword_query": keyword_query,
                "database": database,
                "keyword_matches": [
                    km.to_json_serializable() for km in keyword_matches
                ],
                "query_matches": [
                    qm.to_json_serializable() for qm in query_matches
                ],
                "candidate_join_networks": [
                    cjn.to_json_serializable() for cjn in cjns
                ],
                "keyword_match_count": len(keyword_matches),
                "query_match_count": len(query_matches),
                "candidate_join_network_count": len(cjns),
                "generated_query": generated_query,
            }

            polly_result = PollyResult(
                **database_result, index_handler=index_handler
            )

            result.append((database_result, polly_result))

        return result

    def _generate_keyword_matches(self, query: str, database: str):
        keyword_matches = LLMKMatcher(
            keyword_query=query,
            database=database,
            llm_chat=self.llm_chat,
        ).generate_keyword_matches()

        return keyword_matches

    def _generate_query_matches(
        self, keyword_matches: List[KeywordMatch] = None
    ):
        pass

    def _init_source_handlers(self):
        for config in self.config.sources_configs:
            handler = SourceHandler(
                config.name,
                config.type,
                config.connection_config,
            )

            self.source_handlers[config.connection_config.database] = handler

            # self._generate_data_source_description(handler)

    def _init_index_handlers(self):
        for source in self.source_handlers:
            index_handler = IndexHandler(
                self.config, self.source_handlers[source]
            )

            if not index_handler.index_exists():
                index_handler.create_index()
                if self.source_handlers[source].type == DataSourceType.DOCUMENT:
                    self.source_handlers[
                        source
                    ].get_collections_and_attributes()

            index_handler.load_schema_graph()

            self.index_handlers[source] = index_handler

    def _generate_data_source_description(self, handler: SourceHandler):
        llm_schema_describer = LLMSchemaDescriber(self.api_key)

        self.logger.info(
            f"Generating data source description for {handler.database} database ({handler.type.value})"
        )

        Path("tmp/").mkdir(exist_ok=True)

        llm_schema_describer.generate_schema_description(handler)

    def execute_query(self, database: str, query):
        print(f"Executing query for database {database}...")
        print("Query:")
        pp(query)

        handler = self.source_handlers[database]
        result = handler.execute_query(query)

        print("\nResult:")
        pp(result)



import re
from string import ascii_lowercase, ascii_uppercase, punctuation

from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import word_tokenize


class Tokenizer:
    stop_words = set(stopwords.words("english")) - {"will"}
    re_spaces = re.compile(r"\s+")
    translate_table = str.maketrans(
        ascii_uppercase + punctuation, ascii_lowercase + " " * len(punctuation)
    )
    wlm = WordNetLemmatizer()
    del translate_table[ord("_")]

    def __init__(self, **kwargs):
        self.tokenize_method = kwargs.get("tokenize_method", "simple")

    def compound_keywords(self, keyword_query):
        return [
            self.tokenize(segment) for segment in keyword_query.split('"')[1::2]
        ]

    def keywords(self, keyword_query):
        return self.tokenize(keyword_query)

    def text_translator(self, text):
        keywords = []
        split_text = text.split(" ")
        for word in split_text:
            # translate_table = word.maketrans(punctuation, len(punctuation) * ' ')
            # word = word.translate(translate_table)
            # word = word.strip()
            word = word.replace(",", "")
            # word = word.replace('.', '')
            # word = word.strip('.')
            word = word.strip()
            keywords.append(word.lower())

        return keywords

    def tokenize(self, text):
        if self.tokenize_method == "simple":
            return [
                keyword
                for keyword in self.text_translator(str(text))
                if keyword not in Tokenizer.stop_words
            ]

        if self.tokenize_method == "nltk":
            word_tokens = word_tokenize(text)
            keywords = [w for w in word_tokens if not w in self.stop_words]
            return keywords

    def extract_keywords(self, keyword_query):
        compound_operators_indexes = []
        valid_keyword_query = []
        compound_keyword_positions = []

        keywords = keyword_query.split(" ")
        keyword_query_length = len(keywords)

        for keyword_position, keyword in enumerate(keywords):
            keyword = keyword.lower()
            keyword = keyword.replace(",", "")
            keyword = keyword.strip()

            if keyword in Tokenizer.stop_words:
                continue

            if keyword == "+":
                if (keyword_position - 1) in compound_keyword_positions:
                    raise Exception(
                        "Unable to process query due to multiple operands for compound operator"
                    )

                if (keyword_position - 1) < 0 or (keyword_position + 1) > (
                    keyword_query_length - 1
                ):
                    raise Exception("Missing operand for compound operator")

                compound_operators_indexes.append(
                    keyword_position,
                )

                valid_keyword_query.append(
                    [
                        keywords[keyword_position - 1],
                        keywords[keyword_position + 1],
                    ],
                )

                compound_keyword_positions.extend(
                    [keyword_position - 1, keyword_position + 1],
                )

                continue

            if keyword_position not in compound_keyword_positions:
                if (keyword_position + 1) < keyword_query_length and keywords[
                    keyword_position + 1
                ] == "+":
                    continue

                valid_keyword_query.append(
                    keyword,
                )

        return valid_keyword_query



import json
import pickle
from pathlib import Path
from pprint import pp
from typing import Generator, List, Tuple

import pymongo

from polly.sources.mongo.iter import MongoIter
from polly.sources.mongo.utils.functions import load_schema_graph_from_json
from polly.sources.mongo.utils.traverser import DocumentTraverser


class MongoHandler:
    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        password: str,
        database: str,
        **kwargs,
    ):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.database_client = pymongo.MongoClient(
            host=f"mongodb://{user}:{password}@{host}",
            port=port,
        )

        self.database_batch_cursor_size = kwargs.get(
            "database_batch_cursor_size",
            10000,
        )

    def execute(self, query):
        base_collection, pipeline_sketch = query.build()

        database = self.database_client[self.database]

        mongo_result = database[base_collection].aggregate(
            pipeline_sketch,
            allowDiskUse=True,
        )

        for item in mongo_result:
            pp(item)

    def get_fk_constraints(self):
        fk_constraints = load_schema_graph_from_json(self.database)

        return fk_constraints

    def get_collections_and_attributes(
        self, dataset_name: str, indexable_attributes=None
    ):
        database = self.database_client[dataset_name]
        print(f"Database: {dataset_name}")

        collections_attributes = {}
        collections_structure = {}
        for collection in self.get_collections_names(dataset_name):
            print(f"Collection: {collection}")
            collections_attributes[collection] = set([])
            collections_structure[collection] = {}

            projection_attributes = None

            if indexable_attributes:
                for item in indexable_attributes:
                    if item[0] == collection:
                        projection_attributes = item[1]

            for document in database[collection].find(
                projection=projection_attributes,
                batch_size=self.database_batch_cursor_size,
            ):
                traverser = DocumentTraverser()
                traverser.traverse(document)
                document_attributes = traverser.get_document_attributes().keys()
                for attribute in document_attributes:

                    if attribute != "_id":
                        collections_attributes[collection].add(
                            attribute,
                        )

                attributes_with_types = traverser.get_document_attributes()

                for attribute in attributes_with_types:
                    if attributes_with_types[attribute] != type(None):
                        collections_structure[collection][attribute] = (
                            attributes_with_types[attribute]
                        )

            collections_attributes[collection] = list(
                collections_attributes[collection]
            )

        print(f"Storing {dataset_name} database structure...")
        Path("tmp/").mkdir(exist_ok=True)
        with open(
            "/".join(["tmp", dataset_name + ".pickle"]),
            mode="wb",
        ) as f:
            pickle.dump(collections_structure, f, protocol=3)

        return collections_attributes

    def get_description_example_values(
        self,
    ) -> Generator[Tuple[str, List[Tuple[str, str]]], None, None]:
        database = self.database_client[self.database]

        collections_attributes = {}
        collections_structure = {}

        indexable_attributes = {}
        attributes_filepath = (
            "./indexable_dataset_attributes/{}_attributes.json".format(
                self.database,
            )
        )

        with open(
            attributes_filepath,
            "r",
        ) as indexable_attributes_file:
            content = json.load(indexable_attributes_file)

            for item in content:
                indexable_attributes[item["table"]] = item["attributes"]

        for collection in self.get_collections_names(self.database):
            document_count = 0
            collections_attributes[collection] = set([])
            collections_structure[collection] = {}

            projection_attributes = None

            if indexable_attributes and collection in indexable_attributes:
                projection_attributes = indexable_attributes[collection]

            traverser = DocumentTraverser()

            print(f"Collection: {collection}")
            print(f"Indexable attributes: {projection_attributes}")
            # continue
            print(
                f"Documents for processing: {database[collection].estimated_document_count()}"
            )
            print(projection_attributes)

            for document in database[collection].find(
                projection=projection_attributes,
                batch_size=self.database_batch_cursor_size,
            ):
                traverser.traverse(document)
                document_count += 1

                if document_count % 50000 == 0:
                    print(f"Processed {document_count} documents")
            # print(collection, traverser.get_document_attributes_values())
            yield collection, traverser.convert_document_attributes_values_to_list_of_tuples(),

            # break
            #     document_attributes = traverser.get_document_attributes().keys()
            #     for attribute in document_attributes:

            #         if attribute != "_id":
            #             collections_attributes[collection].add(
            #                 attribute,
            #             )

            #     attributes_with_types = traverser.get_document_attributes()

            #     for attribute in attributes_with_types:
            #         if attributes_with_types[attribute] != type(None):
            #             collections_structure[collection][attribute] = (
            #                 attributes_with_types[attribute]
            #             )

            # collections_attributes[collection] = list(
            #     collections_attributes[collection]
            # )

    def iterate_over_keywords(
        self, dataset_name: str, schema_index, indexable_attributes, **kwargs
    ):
        schema_index_attributes = schema_index.tables_attributes(
            self.dataset_name, nested_structures=True
        )
        print("Schema index attributes", schema_index_attributes)
        if indexable_attributes:
            for attribute in indexable_attributes:
                if attribute not in schema_index_attributes:
                    print(
                        "Attribute {} not located in Schema Index... Re-check and try again.".format(
                            attribute
                        )
                    )
                    return None

        return MongoIter(
            indexable_attributes,
            dataset_name,
            self.database_client,
            self.database_batch_cursor_size,
            **kwargs,
        )

    def exist_results(self, dataset_name: str, query):
        database = self.database_client[dataset_name]
        collection, mongo_query = query.build()

        count = database[collection].aggregate(mongo_query)

        for item in count:
            if item:
                count.close()
                return True

        return False

    def get_databases(self):
        databases = self.database_client.list_database_names()
        databases = [
            database
            for database in databases
            if database not in ["admin", "local", "config"]
        ]

        return databases

    def get_collections_names(self, dataset_name: str):
        database = self.database_client[dataset_name]
        filter = {"name": {"$regex": r"^(?!system\.)"}}

        collection_names = database.list_collection_names(filter=filter)

        return collection_names



import pickle
from json import load as json_load
from pathlib import Path


def modify_retrieved_documents(retrieved_documents):
    modified_documents = []
    while len(retrieved_documents) > 0:
        document = retrieved_documents.pop()
        # print(document)
        if "_id" in document:
            document["mongo_id"] = str(document["_id"])
        del document["_id"]
        modified_documents.append(document)

    return modified_documents


def load_schema_graph_from_json(database):
    PATH = f"./schema_graphs/{database}_schema_graph.json"
    schema_graph_file_path_exists = Path(PATH).exists()
    schema_graph = {}

    if schema_graph_file_path_exists:
        with open(PATH) as f:
            schema_graph = json_load(f)
        return schema_graph

    raise Exception("Schema Graph file not found")


def load_datasets_ids_from_json():
    PATH = "./collections_ids/datasets_collection_ids.json"
    datasets_ids_file_exists = Path(PATH).is_file()

    if datasets_ids_file_exists:
        with open(PATH) as f:
            return json_load(f)

    raise Exception("Datasets collection IDs file not found")


def load_database_structure_from_file(database_name):
    PATH = f"tmp/{database_name}.pickle"
    with open(PATH, "rb") as f:
        database_structure = pickle.load(
            f,
        )

    return database_structure



from nltk import word_tokenize


def sort_dataframe_by_token_length(df):
    local_df = df.copy()
    local_df["name_length"] = 0
    num_number_col = 0
    for col, dt in local_df.dtypes.items():
        if col == "name_length":
            continue
        if dt == object:
            local_df["name_length"] += local_df.apply(
                lambda row: len(word_tokenize(row[col])), axis=1
            )
        else:
            num_number_col += 1
    local_df["name_length"] += num_number_col
    local_df = local_df.sort_values("name_length", ascending=False).drop(
        "name_length", axis=1
    )
    return local_df


def sort_dataframe_by_bow_size(df):
    local_df = df.copy()
    local_df["bow"] = local_df.apply(lambda row: set(), axis=1)
    num_number_col = 0
    for col, dt in local_df.dtypes.items():
        if col == "bow" or col == "bow_length":
            continue
        if dt == object:
            local_df["bow"] = local_df.apply(
                lambda row: row["bow"] | set(word_tokenize(row[col])), axis=1
            )
        else:
            num_number_col += 1
    local_df["bow_lenght"] = local_df.apply(
        lambda row: len(row["bow"]) + num_number_col, axis=1
    )
    local_df = local_df.sort_values("bow_lenght", ascending=False).drop(
        ["bow_lenght", "bow"], axis=1
    )
    return local_df



import logging

from polly.config import ConfigHandler


def get_logger(name, config: ConfigHandler):
    logging.basicConfig(level=config.logging_mode)

    logger = logging.getLogger(name)
    return logger



import itertools
from typing import List

from polly.core.classes.keyword_match import KeywordMatch
from polly.core.classes.query_match import QueryMatch


class QueryMatchHandler:

    def __init__(self):
        return

    def generate_query_matches(self, keywords, keyword_matches, **kwargs):
        # Input:  A keyword query Q, The set of non-empty non-free tuple-sets Rq
        # Output: The set Mq of query matches for Q
        max_qm_size = kwargs.get("max_qm_size", 3)
        max_merged_qm_size = kwargs.get("max_merged_qm_size", 6)
        query_matches = []

        for i in range(1, min(len(keywords), max_qm_size) + 1):
            for candidate_query_match in itertools.combinations(
                keyword_matches, i
            ):
                if self.has_minimal_cover(candidate_query_match, keywords):
                    merged_query_match = self.merge_schema_filters(
                        candidate_query_match
                    )
                    query_match = QueryMatch(merged_query_match)
                    if len(merged_query_match) <= max_merged_qm_size:
                        query_matches.append(query_match)

        return query_matches

    def has_minimal_cover(self, candidate_query_match, keywords):
        # Input:  A subset CM (Candidate Query Match) to be checked as total and minimal cover
        # Output: If the match candidate is a TOTAL and MINIMAL cover

        # Check whether it is total

        if len(
            {
                keyword
                for keyword_match in candidate_query_match
                for keyword in keyword_match.keywords()
            }
        ) != len(set(keywords)):
            return False

        # Check whether it is minimal
        for element in candidate_query_match:
            if len(
                {
                    keyword
                    for keyword_match in candidate_query_match
                    for keyword in keyword_match.keywords()
                    if keyword_match != element
                }
            ) == len(set(keywords)):
                return False

        return True

    def merge_schema_filters(self, candidate_query_match: List[KeywordMatch]):
        table_hash = {}
        database = ""

        for keyword_match in candidate_query_match:
            joint_schema_filter, value_keyword_matches = table_hash.setdefault(
                keyword_match.table, ({}, set())
            )

            for attribute, keywords in keyword_match.schema_filter:
                joint_schema_filter.setdefault(attribute, set()).update(
                    keywords
                )
                database = keyword_match.database

            if len(keyword_match.value_filter) > 0:
                value_keyword_matches.add(keyword_match)

        merged_qm = set()
        for table, (
            joint_schema_filter,
            value_keyword_matches,
        ) in table_hash.items():
            database = ""
            if len(value_keyword_matches) > 0:
                database = list(value_keyword_matches)[0].database
                joint_value_filter = {
                    attribute: keywords
                    for attribute, keywords in value_keyword_matches.pop().value_filter
                }
            else:
                joint_value_filter = {}

            joint_keyword_match = KeywordMatch(
                table,
                database,
                value_filter=joint_value_filter,
                schema_filter=joint_schema_filter,
            )

            merged_qm.add(joint_keyword_match)
            merged_qm.update(value_keyword_matches)

        return merged_qm
